# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t


def parse_bsub(output: str) -> str:
    """Parse bsub output and return job id.

    :param output: stdout of bsub command
    :returns: job id
    """
    for line in output.split("\n"):
        if line.startswith("Job"):
            return line.split()[1][1:-1]
    return ""


def parse_bsub_error(output: str) -> str:
    """Parse and return error output of a failed bsub command.

    :param output: stderr of qsub command
    :returns: error message
    """
    # Search for first non-empty line
    error_lines = []
    copy_lines = False
    for line in output.split("\n"):
        if line.strip().startswith("**"):
            copy_lines = True
        if line.strip().startswith("-----------"):
            copy_lines = False
        if copy_lines and line.lstrip("*").strip():
            error_lines += [line.lstrip("*").strip()]

    if error_lines:
        return "\n".join(error_lines).strip()

    # If the error message was not parsable, return it as is
    if output.strip():
        return output

    # If output is empty, present standard error
    base_err = "LSF run error"
    return base_err


def parse_jslist_stepid(output: str, step_id: str) -> t.Tuple[str, t.Optional[str]]:
    """Parse and return output of the jslist command run with
    options to obtain step status

    :param output: output of the bjobs command
    :param step_id: allocation id or job step id
    :return: status and return code
    """
    result: t.Tuple[str, t.Optional[str]] = ("NOTFOUND", None)

    for line in output.split("\n"):
        fields = line.split()
        if len(fields) >= 7:
            if fields[0] == step_id:
                stat = fields[6]
                return_code = fields[5]
                result = (stat, return_code)
                break
    return result


def parse_bjobs_jobid(output: str, job_id: str) -> str:
    """Parse and return output of the bjobs command run with options
    to obtain job status.

    :param output: output of the bjobs command
    :param job_id: allocation id or job step id
    :return: status
    """
    result = "NOTFOUND"
    for line in output.split("\n"):
        fields = line.split()
        if len(fields) >= 3:
            if fields[0] == job_id:
                stat = fields[2]
                result = stat
                break
    return result


def parse_bjobs_nodes(output: str) -> t.List[str]:
    """Parse and return the bjobs command run with
    options to obtain node list, i.e. with `-w`.

    This function parses and returns the nodes of
    a job in a list with the duplicates removed.

    :param output: output of the `bjobs -w` command
    :return: compute nodes of the allocation or job
    """
    nodes = []

    lines = output.split("\n")
    nodes_str = lines[1].split()[5]
    nodes = nodes_str.split(":")

    return list(dict.fromkeys(nodes))


def parse_max_step_id_from_jslist(output: str) -> t.Optional[str]:
    """Parse and return the maximum step id from a jslist command.
    This function must be called immedietaly after a call to jsrun,
    and before the next one, to ensure the id of the last spawned task is
    properly returned

    :param output: output bjobs
    :param step_name: the name of the step to query
    :return: the step_id
    """
    max_step_id = None

    for line in output.split("\n"):
        if line.startswith("="):
            continue
        fields = line.split()
        if len(fields) >= 7:
            if fields[0].isdigit():
                if (max_step_id is None) or (int(fields[0]) > max_step_id):
                    max_step_id = int(fields[0])

    if max_step_id:
        return str(max_step_id)
    return None



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
import logging
import pathlib
import time
import typing as t
from dataclasses import dataclass, field

from smartsim._core.control.job import JobEntity

logger = logging.getLogger("TelemetryMonitor")


@dataclass
class Run:
    """
    A Run contains the collection of entities created when a `SmartSim`
    driver script executes `Experiment.start`"""

    timestamp: int
    """the timestamp at the time the `Experiment.start` is called"""
    models: t.List[JobEntity]
    """models started in this run"""
    orchestrators: t.List[JobEntity]
    """orchestrators started in this run"""
    ensembles: t.List[JobEntity]
    """ensembles started in this run"""

    def flatten(
        self, filter_fn: t.Optional[t.Callable[[JobEntity], bool]] = None
    ) -> t.Sequence[JobEntity]:
        """Flatten all `JobEntity`'s in the `Run` into a 1-dimensional list

        :param filter_fn: optional boolean filter that returns
        True for entities to include in the result
        """
        entities = self.models + self.orchestrators + self.ensembles
        if filter_fn:
            entities = [entity for entity in entities if filter_fn(entity)]
        return entities

    @staticmethod
    def load_entity(
        entity_type: str,
        entity_dict: t.Dict[str, t.Any],
        exp_dir: pathlib.Path,
        raw_experiment: t.Dict[str, t.Any],
    ) -> t.List[JobEntity]:
        """Map entity data persisted in a manifest file to an object

        :param entity_type: type of the associated `SmartSimEntity`
        :param entity_dict: raw dictionary deserialized from entity in manifest JSON
        :param exp_dir: root path to experiment outputs
        :param raw_experiment: raw experiment deserialized from manifest JSON
        :return: list of loaded `JobEntity` instances
        """
        entities = []

        # an entity w/parent keys must create entities for the items that it
        # comprises. traverse the children and create each entity
        parent_keys = {"shards", "models"}
        parent_keys = parent_keys.intersection(entity_dict.keys())
        if parent_keys:
            container = "shards" if "shards" in parent_keys else "models"
            child_type = "orchestrator" if container == "shards" else "model"
            for child_entity in entity_dict[container]:
                entity = JobEntity.from_manifest(
                    child_type, child_entity, str(exp_dir), raw_experiment
                )
                entities.append(entity)

            return entities

        # not a parent type, just create the entity w/the entity_type passed in
        entity = JobEntity.from_manifest(
            entity_type, entity_dict, str(exp_dir), raw_experiment
        )
        entities.append(entity)
        return entities

    @staticmethod
    def load_entities(
        entity_type: str,
        run: t.Dict[str, t.Any],
        exp_dir: pathlib.Path,
        raw_experiment: t.Dict[str, t.Any],
    ) -> t.Dict[str, t.List[JobEntity]]:
        """Map a collection of entity data persisted in a manifest file to an object

        :param entity_type: type of the associated `SmartSimEntity`
        :param run: raw dictionary containing `Run` data deserialized from JSON
        :param exp_dir: root path to experiment outputs
        :param raw_experiment: raw experiment deserialized from manifest JSON
        :return: list of loaded `JobEntity` instances
        """
        persisted: t.Dict[str, t.List[JobEntity]] = {
            "model": [],
            "orchestrator": [],
        }
        for item in run[entity_type]:
            entities = Run.load_entity(entity_type, item, exp_dir, raw_experiment)
            for new_entity in entities:
                persisted[new_entity.type].append(new_entity)

        return persisted

    @staticmethod
    def load_run(
        raw_run: t.Dict[str, t.Any],
        exp_dir: pathlib.Path,
        raw_experiment: t.Dict[str, t.Any],
    ) -> "Run":
        """Map run data persisted in a manifest file to an object

        :param raw_run: raw dictionary containing `Run` data deserialized from JSON
        :param exp_dir: root path to experiment outputs
        :param raw_experiment: raw experiment deserialized from manifest JSON
        :return: populated `Run` instance
        """

        # create an output mapping to hold the deserialized entities
        run_entities: t.Dict[str, t.List[JobEntity]] = {
            "model": [],
            "orchestrator": [],
            "ensemble": [],
        }

        # use the output mapping keys to load all the target
        # entities from the deserialized JSON
        for entity_type in run_entities:
            _entities = Run.load_entities(entity_type, raw_run, exp_dir, raw_experiment)

            # load_entities may return a mapping containing types different from
            # entity_type IF it was a parent entity. Iterate through the keys in
            # the output dictionary and put them in the right place
            for entity_type, new_entities in _entities.items():
                if not new_entities:
                    continue
                run_entities[entity_type].extend(new_entities)

        loaded_run = Run(
            raw_run["timestamp"],
            run_entities["model"],
            run_entities["orchestrator"],
            run_entities["ensemble"],
        )
        return loaded_run


@dataclass
class RuntimeManifest:
    """The runtime manifest holds information about the entities created
    at runtime during a SmartSim Experiment. The runtime manifest differs
    from a standard manifest - it may contain multiple experiment
    executions in a `runs` collection and holds information that is unknown
    at design-time, such as IP addresses of host machines.
    """

    name: str
    """The name of the `Experiment` associated to the `RuntimeManifest`"""
    path: pathlib.Path
    """The path to the `Experiment` working directory"""
    launcher: str
    """The launcher type used by the `Experiment`"""
    runs: t.List[Run] = field(default_factory=list)
    """A `List` of 0 to many `Run` instances"""

    @staticmethod
    def load_manifest(file_path: str) -> t.Optional["RuntimeManifest"]:
        """Load a persisted manifest and return the content

        :param file_path: path to the manifest file to load
        :return: deserialized `RuntimeManifest` if the manifest file is found,
        otherwise None
        """
        manifest_dict: t.Optional[t.Dict[str, t.Any]] = None
        try_count, max_attempts = 1, 5

        # allow multiple read attempts in case the manifest is being
        # written at the time load_manifest is called
        while manifest_dict is None and try_count <= max_attempts:
            source = pathlib.Path(file_path)
            source = source.resolve()
            time.sleep(0.01)  # a tiny sleep avoids reading partially written json

            try:
                if text := source.read_text(encoding="utf-8").strip():
                    manifest_dict = json.loads(text)
            except json.JSONDecodeError as ex:
                print(f"Error loading manifest: {ex}")
                # hack/fix: handle issues reading file before it is fully written
                time.sleep(0.1 * try_count)
            finally:
                try_count += 1

        if not manifest_dict:
            return None

        # if we don't have an experiment, the manifest is malformed
        exp = manifest_dict.get("experiment", None)
        if not exp:
            raise ValueError("Manifest missing required experiment")

        # if we don't have runs, the manifest is malformed
        runs = manifest_dict.get("runs", None)
        if runs is None:
            raise ValueError("Manifest missing required runs")

        exp_dir = pathlib.Path(exp["path"])
        runs = [Run.load_run(raw_run, exp_dir, exp) for raw_run in runs]

        manifest = RuntimeManifest(
            name=exp["name"],
            path=exp_dir,
            launcher=exp["launcher"],
            runs=runs,
        )
        return manifest



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Generation Strategies
import random
import typing as t
from itertools import product


# create permutations of all parameters
# single model if parameters only have one value
def create_all_permutations(
    param_names: t.List[str], param_values: t.List[t.List[str]], _n_models: int = 0
) -> t.List[t.Dict[str, str]]:
    perms = list(product(*param_values))
    all_permutations = []
    for permutation in perms:
        temp_model = dict(zip(param_names, permutation))
        all_permutations.append(temp_model)
    return all_permutations


def step_values(
    param_names: t.List[str], param_values: t.List[t.List[str]], _n_models: int = 0
) -> t.List[t.Dict[str, str]]:
    permutations = []
    for param_value in zip(*param_values):
        permutations.append(dict(zip(param_names, param_value)))
    return permutations


def random_permutations(
    param_names: t.List[str], param_values: t.List[t.List[str]], n_models: int = 0
) -> t.List[t.Dict[str, str]]:
    permutations = create_all_permutations(param_names, param_values)

    # sample from available permutations if n_models is specified
    if n_models and n_models < len(permutations):
        permutations = random.sample(permutations, n_models)

    return permutations



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path as osp
import typing as t
from copy import deepcopy
from os import getcwd

from tabulate import tabulate

from .._core._install.builder import Device
from ..error import (
    EntityExistsError,
    SmartSimError,
    SSUnsupportedError,
    UserStrategyError,
)
from ..log import get_logger
from ..settings.base import BatchSettings, RunSettings
from .dbobject import DBModel, DBScript
from .entity import SmartSimEntity
from .entityList import EntityList
from .model import Model
from .strategies import create_all_permutations, random_permutations, step_values

logger = get_logger(__name__)

StrategyFunction = t.Callable[
    [t.List[str], t.List[t.List[str]], int], t.List[t.Dict[str, str]]
]


class Ensemble(EntityList[Model]):
    """``Ensemble`` is a group of ``Model`` instances that can
    be treated as a reference to a single instance.
    """

    def __init__(
        self,
        name: str,
        params: t.Dict[str, t.Any],
        path: t.Optional[str] = getcwd(),
        params_as_args: t.Optional[t.List[str]] = None,
        batch_settings: t.Optional[BatchSettings] = None,
        run_settings: t.Optional[RunSettings] = None,
        perm_strat: str = "all_perm",
        **kwargs: t.Any,
    ) -> None:
        """Initialize an Ensemble of Model instances.

        The kwargs argument can be used to pass custom input
        parameters to the permutation strategy.

        :param name: name of the ensemble
        :param params: parameters to expand into ``Model`` members
        :param params_as_args: list of params that should be used as command
            line arguments to the ``Model`` member executables and not written
            to generator files
        :param batch_settings: describes settings for ``Ensemble`` as batch workload
        :param run_settings: describes how each ``Model`` should be executed
        :param replicas: number of ``Model`` replicas to create - a keyword
            argument of kwargs
        :param perm_strategy: strategy for expanding ``params`` into
                             ``Model`` instances from params argument
                             options are "all_perm", "step", "random"
                             or a callable function.
        :return: ``Ensemble`` instance
        """
        self.params = params or {}
        self.params_as_args = params_as_args or []
        self._key_prefixing_enabled = True
        self.batch_settings = batch_settings
        self.run_settings = run_settings
        self.replicas: str

        super().__init__(name, str(path), perm_strat=perm_strat, **kwargs)

    @property
    def models(self) -> t.Collection[Model]:
        """An alias for a shallow copy of the ``entities`` attribute"""
        return list(self.entities)

    def _initialize_entities(self, **kwargs: t.Any) -> None:
        """Initialize all the models within the ensemble based
        on the parameters passed to the ensemble and the permutation
        strategy given at init.

        :raises UserStrategyError: if user generation strategy fails
        """
        strategy = self._set_strategy(kwargs.pop("perm_strat"))
        replicas = kwargs.pop("replicas", None)
        self.replicas = replicas

        # if a ensemble has parameters and run settings, create
        # the ensemble and assign run_settings to each member
        if self.params:
            if self.run_settings:
                param_names, params = self._read_model_parameters()

                # Compute all combinations of model parameters and arguments
                n_models = kwargs.get("n_models", 0)
                all_model_params = strategy(param_names, params, n_models)
                if not isinstance(all_model_params, list):
                    raise UserStrategyError(strategy)

                for i, param_set in enumerate(all_model_params):
                    if not isinstance(param_set, dict):
                        raise UserStrategyError(strategy)
                    run_settings = deepcopy(self.run_settings)
                    model_name = "_".join((self.name, str(i)))
                    model = Model(
                        name=model_name,
                        params=param_set,
                        path=osp.join(self.path, model_name),
                        run_settings=run_settings,
                        params_as_args=self.params_as_args,
                    )
                    model.enable_key_prefixing()
                    model.params_to_args()
                    logger.debug(
                        f"Created ensemble member: {model_name} in {self.name}"
                    )
                    self.add_model(model)
            # cannot generate models without run settings
            else:
                raise SmartSimError(
                    "Ensembles without 'params' or 'replicas' argument to "
                    "expand into members cannot be given run settings"
                )
        else:
            if self.run_settings:
                if replicas:
                    for i in range(replicas):
                        model_name = "_".join((self.name, str(i)))
                        model = Model(
                            name=model_name,
                            params={},
                            path=osp.join(self.path, model_name),
                            run_settings=deepcopy(self.run_settings),
                        )
                        model.enable_key_prefixing()
                        logger.debug(
                            f"Created ensemble member: {model_name} in {self.name}"
                        )
                        self.add_model(model)
                else:
                    raise SmartSimError(
                        "Ensembles without 'params' or 'replicas' argument to "
                        "expand into members cannot be given run settings"
                    )
            # if no params, no run settings and no batch settings, error because we
            # don't know how to run the ensemble
            elif not self.batch_settings:
                raise SmartSimError(
                    "Ensemble must be provided batch settings or run settings"
                )
            else:
                logger.info("Empty ensemble created for batch launch")

    def add_model(self, model: Model) -> None:
        """Add a model to this ensemble

        :param model: model instance to be added
        :raises TypeError: if model is not an instance of ``Model``
        :raises EntityExistsError: if model already exists in this ensemble
        """
        if not isinstance(model, Model):
            raise TypeError(
                f"Argument to add_model was of type {type(model)}, not Model"
            )
        # "in" operator uses model name for __eq__
        if model in self.entities:
            raise EntityExistsError(
                f"Model {model.name} already exists in ensemble {self.name}"
            )

        if self._db_models:
            self._extend_entity_db_models(model, self._db_models)
        if self._db_scripts:
            self._extend_entity_db_scripts(model, self._db_scripts)

        self.entities.append(model)

    def register_incoming_entity(self, incoming_entity: SmartSimEntity) -> None:
        """Register future communication between entities.

        Registers the named data sources that this entity
        has access to by storing the key_prefix associated
        with that entity

        Only python clients can have multiple incoming connections

        :param incoming_entity: The entity that data will be received from
        """
        for model in self.models:
            model.register_incoming_entity(incoming_entity)

    def enable_key_prefixing(self) -> None:
        """If called, each model within this ensemble will prefix its key with its
        own model name.
        """
        for model in self.models:
            model.enable_key_prefixing()

    def query_key_prefixing(self) -> bool:
        """Inquire as to whether each model within the ensemble will prefix their keys

        :returns: True if all models have key prefixing enabled, False otherwise
        """
        return all(model.query_key_prefixing() for model in self.models)

    def attach_generator_files(
        self,
        to_copy: t.Optional[t.List[str]] = None,
        to_symlink: t.Optional[t.List[str]] = None,
        to_configure: t.Optional[t.List[str]] = None,
    ) -> None:
        """Attach files to each model within the ensemble for generation

        Attach files needed for the entity that, upon generation,
        will be located in the path of the entity.

        During generation, files "to_copy" are copied into
        the path of the entity, and files "to_symlink" are
        symlinked into the path of the entity.

        Files "to_configure" are text based model input files where
        parameters for the model are set. Note that only models
        support the "to_configure" field. These files must have
        fields tagged that correspond to the values the user
        would like to change. The tag is settable but defaults
        to a semicolon e.g. THERMO = ;10;

        :param to_copy: files to copy
        :param to_symlink: files to symlink
        :param to_configure: input files with tagged parameters
        """
        for model in self.models:
            model.attach_generator_files(
                to_copy=to_copy, to_symlink=to_symlink, to_configure=to_configure
            )

    @property
    def attached_files_table(self) -> str:
        """Return a plain-text table with information about files
        attached to models belonging to this ensemble.

        :returns: A table of all files attached to all models
        """
        if not self.models:
            return "The ensemble is empty, no files to show."

        table = tabulate(
            [[model.name, model.attached_files_table] for model in self.models],
            headers=["Model name", "Files"],
            tablefmt="grid",
        )

        return table

    def print_attached_files(self) -> None:
        """Print table of attached files to std out"""
        print(self.attached_files_table)

    @staticmethod
    def _set_strategy(strategy: str) -> StrategyFunction:
        """Set the permutation strategy for generating models within
        the ensemble

        :param strategy: name of the strategy or callable function
        :raises SSUnsupportedError: if str name is not supported
        :return: strategy function
        """
        if strategy == "all_perm":
            return create_all_permutations
        if strategy == "step":
            return step_values
        if strategy == "random":
            return random_permutations
        if callable(strategy):
            return strategy
        raise SSUnsupportedError(
            f"Permutation strategy given is not supported: {strategy}"
        )

    def _read_model_parameters(self) -> t.Tuple[t.List[str], t.List[t.List[str]]]:
        """Take in the parameters given to the ensemble and prepare to
        create models for the ensemble

        :raises TypeError: if params are of the wrong type
        :return: param names and values for permutation strategy
        """

        if not isinstance(self.params, dict):
            raise TypeError(
                "Ensemble initialization argument 'params' must be of type dict"
            )

        param_names: t.List[str] = []
        parameters: t.List[t.List[str]] = []
        for name, val in self.params.items():
            param_names.append(name)

            if isinstance(val, list):
                val = [str(v) for v in val]
                parameters.append(val)
            elif isinstance(val, (int, str)):
                parameters.append([str(val)])
            else:
                raise TypeError(
                    "Incorrect type for ensemble parameters\n"
                    + "Must be list, int, or string."
                )
        return param_names, parameters

    def add_ml_model(
        self,
        name: str,
        backend: str,
        model: t.Optional[bytes] = None,
        model_path: t.Optional[str] = None,
        device: str = Device.CPU.value.upper(),
        devices_per_node: int = 1,
        first_device: int = 0,
        batch_size: int = 0,
        min_batch_size: int = 0,
        min_batch_timeout: int = 0,
        tag: str = "",
        inputs: t.Optional[t.List[str]] = None,
        outputs: t.Optional[t.List[str]] = None,
    ) -> None:
        """A TF, TF-lite, PT, or ONNX model to load into the DB at runtime

        Each ML Model added will be loaded into an
        orchestrator (converged or not) prior to the execution
        of every entity belonging to this ensemble

        One of either model (in memory representation) or model_path (file)
        must be provided

        :param name: key to store model under
        :param model: model in memory
        :param model_path: serialized model
        :param backend: name of the backend (TORCH, TF, TFLITE, ONNX)
        :param device: name of device for execution
        :param devices_per_node: number of GPUs per node in multiGPU nodes
        :param first_device: first device in multi-GPU nodes to use for execution,
                             defaults to 0; ignored if devices_per_node is 1
        :param batch_size: batch size for execution
        :param min_batch_size: minimum batch size for model execution
        :param min_batch_timeout: time to wait for minimum batch size
        :param tag: additional tag for model information
        :param inputs: model inputs (TF only)
        :param outputs: model outupts (TF only)
        """
        db_model = DBModel(
            name=name,
            backend=backend,
            model=model,
            model_file=model_path,
            device=device,
            devices_per_node=devices_per_node,
            first_device=first_device,
            batch_size=batch_size,
            min_batch_size=min_batch_size,
            min_batch_timeout=min_batch_timeout,
            tag=tag,
            inputs=inputs,
            outputs=outputs,
        )
        dupe = next(
            (
                db_model.name
                for ensemble_ml_model in self._db_models
                if ensemble_ml_model.name == db_model.name
            ),
            None,
        )
        if dupe:
            raise SSUnsupportedError(
                f'An ML Model with name "{db_model.name}" already exists'
            )
        self._db_models.append(db_model)
        for entity in self.models:
            self._extend_entity_db_models(entity, [db_model])

    def add_script(
        self,
        name: str,
        script: t.Optional[str] = None,
        script_path: t.Optional[str] = None,
        device: str = Device.CPU.value.upper(),
        devices_per_node: int = 1,
        first_device: int = 0,
    ) -> None:
        """TorchScript to launch with every entity belonging to this ensemble

        Each script added to the model will be loaded into an
        orchestrator (converged or not) prior to the execution
        of every entity belonging to this ensemble

        Device selection is either "GPU" or "CPU". If many devices are
        present, a number can be passed for specification e.g. "GPU:1".

        Setting ``devices_per_node=N``, with N greater than one will result
        in the model being stored in the first N devices of type ``device``.

        One of either script (in memory string representation) or script_path (file)
        must be provided

        :param name: key to store script under
        :param script: TorchScript code
        :param script_path: path to TorchScript code
        :param device: device for script execution
        :param devices_per_node: number of devices on each host
        :param first_device: first device to use on each host
        """
        db_script = DBScript(
            name=name,
            script=script,
            script_path=script_path,
            device=device,
            devices_per_node=devices_per_node,
            first_device=first_device,
        )
        dupe = next(
            (
                db_script.name
                for ensemble_script in self._db_scripts
                if ensemble_script.name == db_script.name
            ),
            None,
        )
        if dupe:
            raise SSUnsupportedError(
                f'A Script with name "{db_script.name}" already exists'
            )
        self._db_scripts.append(db_script)
        for entity in self.models:
            self._extend_entity_db_scripts(entity, [db_script])

    def add_function(
        self,
        name: str,
        function: t.Optional[str] = None,
        device: str = Device.CPU.value.upper(),
        devices_per_node: int = 1,
        first_device: int = 0,
    ) -> None:
        """TorchScript function to launch with every entity belonging to this ensemble

        Each script function to the model will be loaded into a
        non-converged orchestrator prior to the execution
        of every entity belonging to this ensemble.

        For converged orchestrators, the :meth:`add_script` method should be used.

        Device selection is either "GPU" or "CPU". If many devices are
        present, a number can be passed for specification e.g. "GPU:1".

        Setting ``devices_per_node=N``, with N greater than one will result
        in the script being stored in the first N devices of type ``device``;
        alternatively, setting ``first_device=M`` will result in the script
        being stored on nodes M through M + N - 1.

        :param name: key to store function under
        :param function: TorchScript code
        :param device: device for script execution
        :param devices_per_node: number of devices on each host
        :param first_device: first device to use on each host
        """
        db_script = DBScript(
            name=name,
            script=function,
            device=device,
            devices_per_node=devices_per_node,
            first_device=first_device,
        )
        dupe = next(
            (
                db_script.name
                for ensemble_script in self._db_scripts
                if ensemble_script.name == db_script.name
            ),
            None,
        )
        if dupe:
            raise SSUnsupportedError(
                f'A Script with name "{db_script.name}" already exists'
            )
        self._db_scripts.append(db_script)
        for entity in self.models:
            self._extend_entity_db_scripts(entity, [db_script])

    @staticmethod
    def _extend_entity_db_models(model: Model, db_models: t.List[DBModel]) -> None:
        """
        Ensures that the Machine Learning model names being added to the Ensemble
        are unique.

        This static method checks if the provided ML model names already exist in
        the Ensemble. An SSUnsupportedError is raised if any duplicate names are
        found. Otherwise, it appends the given list of DBModels to the Ensemble.

        :param model: SmartSim Model object.
        :param db_models: List of DBModels to append to the Ensemble.
        """
        for add_ml_model in db_models:
            dupe = next(
                (
                    db_model.name
                    for db_model in model.db_models
                    if db_model.name == add_ml_model.name
                ),
                None,
            )
            if dupe:
                raise SSUnsupportedError(
                    f'An ML Model with name "{add_ml_model.name}" already exists'
                )
            model.add_ml_model_object(add_ml_model)

    @staticmethod
    def _extend_entity_db_scripts(model: Model, db_scripts: t.List[DBScript]) -> None:
        """
        Ensures that the script/function names being added to the Ensemble are unique.

        This static method checks if the provided script/function names already exist
        in the Ensemble. An SSUnsupportedError is raised if any duplicate names
        are found. Otherwise, it appends the given list of DBScripts to the
        Ensemble.

        :param model: SmartSim Model object.
        :param db_scripts: List of DBScripts to append to the Ensemble.
        """
        for add_script in db_scripts:
            dupe = next(
                (
                    add_script.name
                    for db_script in model.db_scripts
                    if db_script.name == add_script.name
                ),
                None,
            )
            if dupe:
                raise SSUnsupportedError(
                    f'A Script with name "{add_script.name}" already exists'
                )
            model.add_script_object(add_script)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import subprocess
import typing as t

from smartsim._core.config import CONFIG


def configure_parser(parser: argparse.ArgumentParser) -> None:
    """Builds the parser for the command"""
    parser.add_argument(
        "--dragon",
        action="store_true",
        default=False,
        help="Terminate Dragon environment resources if"
        "any remain after experiment completion",
    )


def _do_dragon_teardown() -> int:
    """Run dragon-cleanup script to destroy all remaining dragon resources"""
    env = os.environ.copy()
    dragon_cleanup = next(CONFIG.core_path.rglob("dragon-cleanup"), None)
    if dragon_cleanup is None:
        print("dragon-cleanup not found. Skipping cleanup")
        return 0

    # Use popen to avoid `dragon-cleanup` doing a kill on all
    # python processes, terminating `smart teardown`, and the
    # subprocess handling `dragon-cleanup`. Child processes using
    # subprocess.run are killed and cleanup is interrupted
    with subprocess.Popen(
        [str(dragon_cleanup.absolute())],
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        process.wait()
        return process.returncode


def execute(
    args: argparse.Namespace, _unparsed_args: t.Optional[t.List[str]] = None, /
) -> int:
    if args.dragon:
        return _do_dragon_teardown()

    return 0



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import time
import typing as t
from subprocess import PIPE
from threading import RLock

import psutil

from ...error import LauncherError
from ...log import ContextThread, get_logger
from ..utils.helpers import check_dev_log_level
from .util.shell import execute_async_cmd, execute_cmd

logger = get_logger(__name__)
VERBOSE_TM = check_dev_log_level()  # pylint: disable=invalid-name

TM_INTERVAL = 1


class TaskManager:
    """The Task Manager watches the subprocesses launched through
    the asyncronous shell interface. Each task is a wrapper
    around the Popen/Process instance.

    The Task Managers polls processes on TM_INTERVAL
    and detects job failure and completion. Upon termination, the
    task returncode, output, and error are added to the task history.

    When a launcher uses the task manager to start a task, the task
    is either managed (by a WLM) or unmanaged (meaning not managed by
    a WLM). In the latter case, the Task manager is responsible for the
    lifecycle of the process.
    """

    def __init__(self) -> None:
        """Initialize a task manager thread."""
        self.actively_monitoring = False
        self.task_history: t.Dict[
            str, t.Tuple[t.Optional[int], t.Optional[str], t.Optional[str]]
        ] = {}
        self.tasks: t.List[Task] = []
        self._lock = RLock()

    def start(self) -> None:
        """Start the task manager thread

        The TaskManager is run as a daemon thread meaning
        that it will die when the main thread dies.
        """
        monitor = ContextThread(name="TaskManager", daemon=True, target=self.run)
        monitor.start()

    def run(self) -> None:
        """Start monitoring Tasks"""

        if VERBOSE_TM:
            logger.debug("Starting Task Manager")

        self.actively_monitoring = True
        while self.actively_monitoring:
            time.sleep(TM_INTERVAL)

            for task in self.tasks:
                returncode = task.check_status()  # poll and set returncode
                # has to be != None because returncode can be 0
                if returncode is not None:
                    output, error = task.get_io()
                    self.add_task_history(task.pid, returncode, output, error)
                    self.remove_task(task.pid)

            if len(self) == 0:
                self.actively_monitoring = False
                if VERBOSE_TM:
                    logger.debug("Sleeping, no tasks to monitor")

    def start_task(
        self,
        cmd_list: t.List[str],
        cwd: str,
        env: t.Optional[t.Dict[str, str]] = None,
        out: int = PIPE,
        err: int = PIPE,
    ) -> str:
        """Start a task managed by the TaskManager

        This is an "unmanaged" task, meaning it is NOT managed
        by a workload manager

        :param cmd_list: command to run
        :param cwd: current working directory
        :param env: environment to launch with
        :param out: output file
        :param err: error file
        :return: task id
        """
        with self._lock:
            proc = execute_async_cmd(cmd_list, cwd, env=env, out=out, err=err)
            task = Task(proc)
            if VERBOSE_TM:
                logger.debug(f"Starting Task {task.pid}")
            self.tasks.append(task)
            self.task_history[task.pid] = (None, None, None)
            return task.pid

    @staticmethod
    def start_and_wait(
        cmd_list: t.List[str],
        cwd: str,
        env: t.Optional[t.Dict[str, str]] = None,
        timeout: t.Optional[int] = None,
    ) -> t.Tuple[int, str, str]:
        """Start a task not managed by the TaskManager

        This method is used by launchers to launch managed tasks
        meaning that they ARE managed by a WLM.

        This is primarily used for batch job launches

        :param cmd_list: command to run
        :param cwd: current working directory
        :param env: environment to launch with
        :param timeout: time to wait
        :return: returncode, output, and err
        """
        returncode, out, err = execute_cmd(cmd_list, cwd=cwd, env=env, timeout=timeout)
        if VERBOSE_TM:
            logger.debug("Ran and waited on task")
        return returncode, out, err

    def add_existing(self, task_id: int) -> None:
        """Add existing task to be managed by the TaskManager

        :param task_id: task id of existing task
        :raises LauncherError: If task cannot be found
        """
        with self._lock:
            try:
                process = psutil.Process(pid=task_id)
                task = Task(process)
                self.tasks.append(task)
                self.task_history[task.pid] = (None, None, None)
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                msg = f"Process provided {task_id} does not exist"
                raise LauncherError(msg) from None

    def remove_task(self, task_id: str) -> None:
        """Remove a task from the TaskManager

        :param task_id: id of the task to remove
        """
        with self._lock:
            if VERBOSE_TM:
                logger.debug(f"Removing Task {task_id}")
            try:
                task = self[task_id]
                if task.is_alive:
                    task.kill()
                    returncode = task.check_status()
                    out, err = task.get_io()
                    self.add_task_history(task_id, returncode, out, err)
                self.tasks.remove(task)
            except psutil.NoSuchProcess:
                logger.debug("Failed to kill a task during removal")
            except KeyError:
                logger.debug("Failed to remove a task, task was already removed")

    def get_task_update(
        self, task_id: str
    ) -> t.Tuple[str, t.Optional[int], t.Optional[str], t.Optional[str]]:
        """Get the update of a task

        :param task_id: task id
        :return: status, returncode, output, error
        """
        with self._lock:
            try:
                return_code, out, err = self.task_history[task_id]
                # has to be == None because rc can be 0
                if return_code is None:
                    try:
                        task = self[task_id]
                        return task.status, return_code, out, err
                    # removed forcefully either by OS or us, no returncode set
                    # either way, job has completed and we won't have returncode
                    # Usually hits when jobs last less then the TM_INTERVAL
                    except (KeyError, psutil.NoSuchProcess):
                        return "Completed", return_code, out, err

                # process has completed, status set manually as we don't
                # save task statuses during runtime.
                else:
                    if return_code != 0:
                        return "Failed", return_code, out, err
                    return "Completed", return_code, out, err
            except KeyError:
                logger.warning(f"Task {task_id} not found in task history dictionary")

        return "Failed", -1, "", ""

    def add_task_history(
        self,
        task_id: str,
        returncode: t.Optional[int] = None,
        out: t.Optional[str] = None,
        err: t.Optional[str] = None,
    ) -> None:
        """Add a task to the task history

        Add a task to record its future returncode, output and error

        :param task_id: id of the task
        :param returncode: returncode
        :param out: output
        :param err: output
        """
        self.task_history[task_id] = (returncode, out, err)

    def __getitem__(self, task_id: str) -> Task:
        with self._lock:
            for task in self.tasks:
                if task.pid == task_id:
                    return task
            raise KeyError

    def __len__(self) -> int:
        with self._lock:
            return len(self.tasks)


class Task:
    def __init__(self, process: psutil.Process) -> None:
        """Initialize a task

        :param process: Popen object
        """
        self.process = process
        self.pid = str(self.process.pid)

    def check_status(self) -> t.Optional[int]:
        """Ping the job and return the returncode if finished

        :return: returncode if finished otherwise None
        """
        if self.owned and isinstance(self.process, psutil.Popen):
            poll_result = self.process.poll()
            if poll_result is not None:
                return int(poll_result)
            return None
        # we can't manage Processed we don't own
        # have to rely on .kill() to stop.
        return self.returncode

    def get_io(self) -> t.Tuple[t.Optional[str], t.Optional[str]]:
        """Get the IO from the subprocess

        :return: output and error from the Popen
        """
        # Process class does not implement communicate
        if not self.owned or not isinstance(self.process, psutil.Popen):
            return None, None

        output, error = self.process.communicate()
        if output:
            output = output.decode("utf-8")
        if error:
            error = error.decode("utf-8")
        return output, error

    def kill(self, timeout: int = 10) -> None:
        """Kill the subprocess and all children"""

        def kill_callback(proc: psutil.Process) -> None:
            logger.debug(f"Process terminated with kill {proc.pid}")

        children = self.process.children(recursive=True)
        children.append(self.process)  # add parent process to be killed

        for child in children:
            child.kill()

        _, alive = psutil.wait_procs(children, timeout=timeout, callback=kill_callback)
        if alive:
            for proc in alive:
                logger.warning(f"Unable to kill emitted process {proc.pid}")

    def terminate(self, timeout: int = 10) -> None:
        """Terminate a this process and all children.

        :param timeout: time to wait for task death
        """

        def terminate_callback(proc: psutil.Process) -> None:
            logger.debug(f"Cleanly terminated task {proc.pid}")

        children = self.process.children(recursive=True)
        children.append(self.process)  # add parent process to be killed

        # try SIGTERM first for clean exit
        for child in children:
            if VERBOSE_TM:
                logger.debug(child)
            child.terminate()

        # wait for termination
        _, alive = psutil.wait_procs(
            children, timeout=timeout, callback=terminate_callback
        )

        if alive:
            logger.debug("SIGTERM failed, using SIGKILL")
            self.process.kill()

    def wait(self) -> None:
        self.process.wait()

    @property
    def returncode(self) -> t.Optional[int]:
        if self.owned and isinstance(self.process, psutil.Popen):
            if self.process.returncode is not None:
                return int(self.process.returncode)
            return None
        if self.is_alive:
            return None
        return 0

    @property
    def is_alive(self) -> bool:
        return self.process.is_running()

    @property
    def status(self) -> str:
        return self.process.status()

    @property
    def owned(self) -> bool:
        if isinstance(self.process, psutil.Popen):
            return True
        return False



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import atexit
import fileinput
import itertools
import json
import os
import subprocess
import sys
import typing as t
from collections import defaultdict
from pathlib import Path
from threading import RLock

import psutil
import zmq
import zmq.auth.thread

from ...._core.launcher.dragon import dragonSockets
from ....error.errors import SmartSimError
from ....log import get_logger
from ...config import get_config
from ...schemas import (
    DragonBootstrapRequest,
    DragonBootstrapResponse,
    DragonHandshakeRequest,
    DragonHandshakeResponse,
    DragonRequest,
    DragonResponse,
    DragonShutdownRequest,
)
from ...utils.network import find_free_port, get_best_interface_and_address

logger = get_logger(__name__)

_SchemaT = t.TypeVar("_SchemaT", bound=t.Union[DragonRequest, DragonResponse])

DRG_LOCK = RLock()


class DragonConnector:
    """This class encapsulates the functionality needed
    to start a Dragon server and communicate with it.
    """

    def __init__(self) -> None:
        self._context: zmq.Context[t.Any] = zmq.Context.instance()
        self._context.setsockopt(zmq.REQ_CORRELATE, 1)
        self._context.setsockopt(zmq.REQ_RELAXED, 1)
        self._authenticator: t.Optional[zmq.auth.thread.ThreadAuthenticator] = None
        config = get_config()
        self._reset_timeout(config.dragon_server_timeout)
        self._dragon_head_socket: t.Optional[zmq.Socket[t.Any]] = None
        self._dragon_head_process: t.Optional[subprocess.Popen[bytes]] = None
        # Returned by dragon head, useful if shutdown is to be requested
        # but process was started by another connector
        self._dragon_head_pid: t.Optional[int] = None
        self._dragon_server_path = config.dragon_server_path
        logger.debug(f"Dragon Server path was set to {self._dragon_server_path}")
        self._env_vars: t.Dict[str, str] = {}
        if self._dragon_server_path is None:
            raise SmartSimError(
                "DragonConnector could not find the dragon server path. "
                "This should not happen if the Connector was started by an "
                "experiment.\nIf the DragonConnector was started manually, "
                "then the environment variable SMARTSIM_DRAGON_SERVER_PATH "
                "should be set to an existing directory."
            )

    @property
    def is_connected(self) -> bool:
        """Whether the Connector established a connection to the server

        :return: True if connected
        """
        return self._dragon_head_socket is not None

    @property
    def can_monitor(self) -> bool:
        """Whether the Connector knows the PID of the dragon server head process
        and can monitor its status

        :return: True if the server can be monitored"""
        return self._dragon_head_pid is not None

    def _handshake(self, address: str) -> None:
        self._dragon_head_socket = dragonSockets.get_secure_socket(
            self._context, zmq.REQ, False
        )
        self._dragon_head_socket.connect(address)
        try:
            dragon_handshake = _assert_schema_type(
                self.send_request(DragonHandshakeRequest()), DragonHandshakeResponse
            )
            self._dragon_head_pid = dragon_handshake.dragon_pid
            logger.debug(
                f"Successful handshake with Dragon server at address {address}"
            )
        except (zmq.ZMQError, zmq.Again) as e:
            logger.debug(e)
            self._dragon_head_socket.close()
            self._dragon_head_socket = None

            raise SmartSimError(
                f"Unsuccessful handshake with Dragon server at address {address}"
            ) from e

    def _reset_timeout(self, timeout: int = get_config().dragon_server_timeout) -> None:
        self._context.setsockopt(zmq.SNDTIMEO, value=timeout)
        self._context.setsockopt(zmq.RCVTIMEO, value=timeout)
        if self._authenticator is not None and self._authenticator.thread is not None:
            try:
                self._authenticator.thread.authenticator.zap_socket.setsockopt(
                    zmq.SNDTIMEO, timeout
                )
                self._authenticator.thread.authenticator.zap_socket.setsockopt(
                    zmq.RCVTIMEO, timeout
                )
            except zmq.ZMQError:
                pass

    def ensure_connected(self) -> None:
        """Ensure that the Connector established a connection to the server

        If the Connector is not connected, attempt to connect and raise an error
        on failure.

        :raises SmartSimError: if connection cannot be established
        """
        if not self.is_connected:
            self.connect_to_dragon()
        if not self.is_connected:
            raise SmartSimError("Could not connect to Dragon server")

    def _get_new_authenticator(
        self, timeout: int = get_config().dragon_server_timeout
    ) -> None:
        if self._authenticator is not None:
            if self._authenticator.thread is not None:
                try:
                    logger.debug("Closing ZAP socket")
                    self._authenticator.thread.authenticator.zap_socket.close()
                except Exception as e:
                    logger.debug(f"Could not close ZAP socket, {e}")
            try:
                self._authenticator.stop()
            except zmq.Again:
                logger.debug("Could not stop authenticator")
        try:
            self._authenticator = dragonSockets.get_authenticator(
                self._context, timeout
            )
            return
        except RuntimeError as e:
            logger.error("Could not get authenticator")
            raise e from None

    @staticmethod
    def _get_dragon_log_level() -> str:
        smartsim_to_dragon = defaultdict(lambda: "NONE")
        smartsim_to_dragon["developer"] = "INFO"
        return smartsim_to_dragon.get(get_config().log_level, "NONE")

    def _connect_to_existing_server(self, path: Path) -> None:
        config = get_config()
        dragon_config_log = path / config.dragon_log_filename

        if not dragon_config_log.is_file():
            return

        dragon_confs = self._parse_launched_dragon_server_info_from_files(
            [dragon_config_log]
        )
        logger.debug(dragon_confs)

        for dragon_conf in dragon_confs:
            logger.debug(
                "Found dragon server config file. Checking if the server"
                f" is still up at address {dragon_conf['address']}."
            )
            try:
                self._reset_timeout()
                self._get_new_authenticator(-1)
                self._handshake(dragon_conf["address"])
            except SmartSimError as e:
                logger.error(e)
            finally:
                self._reset_timeout(config.dragon_server_timeout)
            if self.is_connected:
                logger.debug("Connected to existing Dragon server")
                return

    def _start_connector_socket(self, socket_addr: str) -> zmq.Socket[t.Any]:
        config = get_config()
        connector_socket: t.Optional[zmq.Socket[t.Any]] = None
        self._reset_timeout(config.dragon_server_startup_timeout)
        self._get_new_authenticator(-1)
        connector_socket = dragonSockets.get_secure_socket(self._context, zmq.REP, True)
        logger.debug(f"Binding connector to {socket_addr}")
        connector_socket.bind(socket_addr)
        if connector_socket is None:
            raise SmartSimError("Socket failed to initialize")

        return connector_socket

    def load_persisted_env(self) -> t.Dict[str, str]:
        """Load key-value pairs from a .env file created during dragon installation

        :return: Key-value pairs stored in .env file"""
        if self._env_vars:
            # use previously loaded env vars.
            return self._env_vars

        config = get_config()

        if not config.dragon_dotenv.exists():
            self._env_vars = {}
            return self._env_vars

        with open(config.dragon_dotenv, encoding="utf-8") as dot_env:
            for kvp in dot_env.readlines():
                split = kvp.strip().split("=", maxsplit=1)
                key, value = split[0], split[-1]
                self._env_vars[key] = value

        return self._env_vars

    def merge_persisted_env(self, current_env: t.Dict[str, str]) -> t.Dict[str, str]:
        """Combine the current environment variable set with the dragon .env by adding
        Dragon-specific values and prepending any new values to existing keys

        :param current_env: Environment which has to be merged with .env variables
        :return: Merged environment
        """
        # ensure we start w/a complete env from current env state
        merged_env: t.Dict[str, str] = {**current_env}

        # copy all the values for dragon straight into merged_env
        merged_env.update(
            {k: v for k, v in self._env_vars.items() if k.startswith("DRAGON")}
        )

        # prepend dragon env updates into existing env vars
        for key, value in self._env_vars.items():
            if not key.startswith("DRAGON"):
                if current_value := current_env.get(key, None):
                    # when a key is not dragon specific, don't overwrite the current
                    # value. instead, prepend the value dragon needs to/current env
                    value = f"{value}:{current_value}"
                merged_env[key] = value
        return merged_env

    def connect_to_dragon(self) -> None:
        """Connect to Dragon server

        :raises SmartSimError: If connection cannot be established
        """
        config = get_config()
        with DRG_LOCK:
            # TODO use manager instead
            if self.is_connected:
                return
            if self._dragon_server_path is None:
                raise SmartSimError("Path to Dragon server not set.")

            logger.info(
                "Establishing connection with Dragon server or starting a new one..."
            )

            path = _resolve_dragon_path(self._dragon_server_path)

            self._connect_to_existing_server(path)
            if self.is_connected:
                return

            path.mkdir(parents=True, exist_ok=True)

            local_address = get_best_interface_and_address().address
            if local_address is None:
                # TODO parse output file
                raise SmartSimError(
                    "Could not determine SmartSim's local address, "
                    "the Dragon server could not be started."
                )
            # find first available port >= 5995
            port = find_free_port(start=5995)
            socket_addr = f"tcp://{local_address}:{port}"
            connector_socket = self._start_connector_socket(socket_addr)

            cmd = [
                "dragon",
                "-t",
                config.dragon_transport,
                "-l",
                DragonConnector._get_dragon_log_level(),
                sys.executable,
                "-m",
                "smartsim._core.entrypoints.dragon",
                "+launching_address",
                socket_addr,
            ]

            dragon_out_file = path / "dragon_head.out"
            dragon_err_file = path / "dragon_head.err"

            self.load_persisted_env()
            merged_env = self.merge_persisted_env(os.environ.copy())
            merged_env.update({"PYTHONUNBUFFERED": "1"})

            with (
                open(dragon_out_file, "w", encoding="utf-8") as dragon_out,
                open(dragon_err_file, "w", encoding="utf-8") as dragon_err,
            ):
                logger.debug(f"Starting Dragon environment: {' '.join(cmd)}")

                # pylint: disable-next=consider-using-with
                self._dragon_head_process = subprocess.Popen(
                    args=cmd,
                    bufsize=0,
                    stderr=dragon_err.fileno(),
                    stdout=dragon_out.fileno(),
                    cwd=path,
                    shell=False,
                    env=merged_env,
                    start_new_session=True,
                )

            server = dragonSockets.as_server(connector_socket)
            logger.debug(f"Listening to {socket_addr}")
            request = _assert_schema_type(server.recv(), DragonBootstrapRequest)
            server.send(
                DragonBootstrapResponse(dragon_pid=self._dragon_head_process.pid)
            )
            connector_socket.close()
            logger.debug(f"Connecting to {request.address}")
            self._reset_timeout(config.dragon_server_timeout)
            self._handshake(request.address)

            # Only the Connector which started the server is
            # responsible of it, that's why we register the
            # cleanup in this code branch.
            # The cleanup function should not have references
            # to this object to avoid Garbage Collector lockup
            server_socket = self._dragon_head_socket
            server_process_pid = self._dragon_head_process.pid

            if server_socket is not None and self._dragon_head_process is not None:
                atexit.register(
                    _dragon_cleanup,
                    server_socket=server_socket,
                    server_process_pid=server_process_pid,
                    server_authenticator=self._authenticator,
                )
            elif self._dragon_head_process is not None:
                self._dragon_head_process.wait(1.0)
                if self._dragon_head_process.stdout:
                    for line in iter(self._dragon_head_process.stdout.readline, b""):
                        logger.info(line.decode("utf-8").rstrip())
                if self._dragon_head_process.stderr:
                    for line in iter(self._dragon_head_process.stderr.readline, b""):
                        logger.warning(line.decode("utf-8").rstrip())
                logger.warning(self._dragon_head_process.returncode)
            else:
                logger.warning("Could not start Dragon server as subprocess")

    def cleanup(self) -> None:
        """Shut down Dragon server and authenticator thread"""
        if self._dragon_head_socket is not None and self._dragon_head_pid is not None:
            _dragon_cleanup(
                server_socket=self._dragon_head_socket,
                server_process_pid=self._dragon_head_pid,
                server_authenticator=self._authenticator,
            )
            self._dragon_head_socket = None
            self._dragon_head_pid = None
            self._authenticator = None

    def send_request(self, request: DragonRequest, flags: int = 0) -> DragonResponse:
        """Send a request to the Dragon server using a secure socket

        :param request: The request to send
        :param flags: 0MQ flags, defaults to 0
        :raises SmartSimError: If not connected to Dragon server
        :return: Response from server
        """
        self.ensure_connected()
        if (socket := self._dragon_head_socket) is None:
            raise SmartSimError("Not connected to Dragon")
        return self._send_req_with_socket(socket, request, flags)

    @staticmethod
    def _parse_launched_dragon_server_info_from_iterable(
        stream: t.Iterable[str], num_dragon_envs: t.Optional[int] = None
    ) -> t.List[t.Dict[str, str]]:
        lines = (line.strip() for line in stream)
        lines = (line for line in lines if line)
        tokenized = (line.split(maxsplit=1) for line in lines)
        tokenized = (tokens for tokens in tokenized if len(tokens) > 1)
        dragon_env_jsons = (
            config_dict
            for first, config_dict in tokenized
            if "DRAGON_SERVER_CONFIG" in first
        )
        dragon_envs = (json.loads(config_dict) for config_dict in dragon_env_jsons)

        dragon_envs = (
            dragon_env for dragon_env in dragon_envs if "address" in dragon_env
        )

        if num_dragon_envs:
            sliced_dragon_envs = itertools.islice(dragon_envs, num_dragon_envs)
            return list(sliced_dragon_envs)
        return list(dragon_envs)

    @classmethod
    def _parse_launched_dragon_server_info_from_files(
        cls,
        file_paths: t.List[t.Union[str, "os.PathLike[str]"]],
        num_dragon_envs: t.Optional[int] = None,
    ) -> t.List[t.Dict[str, str]]:
        with fileinput.FileInput(file_paths) as ifstream:
            dragon_envs = cls._parse_launched_dragon_server_info_from_iterable(
                ifstream, num_dragon_envs
            )

            return dragon_envs

    @staticmethod
    def _send_req_with_socket(
        socket: zmq.Socket[t.Any],
        request: DragonRequest,
        send_flags: int = 0,
        recv_flags: int = 0,
    ) -> DragonResponse:
        client = dragonSockets.as_client(socket)
        with DRG_LOCK:
            logger.debug(f"Sending {type(request).__name__}: {request}")
            client.send(request, send_flags)
            response = client.recv(flags=recv_flags)

            logger.debug(f"Received {type(response).__name__}: {response}")
            return response


def _assert_schema_type(obj: object, typ: t.Type[_SchemaT], /) -> _SchemaT:
    if not isinstance(obj, typ):
        raise TypeError(f"Expected schema of type `{typ}`, but got {type(obj)}")
    return obj


def _dragon_cleanup(
    server_socket: t.Optional[zmq.Socket[t.Any]] = None,
    server_process_pid: t.Optional[int] = 0,
    server_authenticator: t.Optional[zmq.auth.thread.ThreadAuthenticator] = None,
) -> None:
    """Clean up resources used by the launcher.
    :param server_socket: (optional) Socket used to connect to dragon environment
    :param server_process_pid: (optional) Process ID of the dragon entrypoint
    :param server_authenticator: (optional) Authenticator used to secure sockets
    """
    try:
        if server_socket is not None:
            print("Sending shutdown request to dragon environment")
            # pylint: disable-next=protected-access
            DragonConnector._send_req_with_socket(
                server_socket, DragonShutdownRequest(), recv_flags=zmq.NOBLOCK
            )
    except zmq.error.ZMQError as e:
        # Can't use the logger as I/O file may be closed
        if not isinstance(e, zmq.Again):
            print("Could not send shutdown request to dragon server")
            print(f"ZMQ error: {e}", flush=True)
    finally:
        print("Sending shutdown request is complete")

    if server_process_pid and psutil.pid_exists(server_process_pid):
        try:
            _, retcode = os.waitpid(server_process_pid, 0)
            print(
                f"Dragon server process shutdown is complete, return code {retcode}",
                flush=True,
            )
        except Exception as e:
            logger.debug(e)

    try:
        if server_authenticator is not None and server_authenticator.is_alive():
            print("Shutting down ZMQ authenticator")
            server_authenticator.stop()
    except Exception:
        print("Authenticator shutdown error")
    else:
        print("Authenticator shutdown is complete")


def _resolve_dragon_path(fallback: t.Union[str, "os.PathLike[str]"]) -> Path:
    dragon_server_path = get_config().dragon_server_path or os.path.join(
        fallback, ".smartsim", "dragon"
    )
    dragon_server_paths = dragon_server_path.split(":")
    if len(dragon_server_paths) > 1:
        logger.warning(
            "Multiple dragon servers not supported, "
            "will connect to (or start) first server in list."
        )
    return Path(dragon_server_paths[0])



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import typing as t


def parse_qsub(output: str) -> str:
    """Parse qsub output and return job id. For PBS, the
    output is the job id itself.

    :param output: stdout of qsub command
    :returns: job id
    """
    return output


def parse_qsub_error(output: str) -> str:
    """Parse and return error output of a failed qsub command.

    :param output: stderr of qsub command
    :returns: error message
    """
    # look for error first
    for line in output.split("\n"):
        if line.startswith("qsub:"):
            error = line.split(":")[1]
            return error.strip()
    # if no error line, take first line
    for line in output.split("\n"):
        return line.strip()
    # if neither, present a base error message
    base_err = "PBS run error"
    return base_err


def parse_qstat_jobid(output: str, job_id: str) -> t.Optional[str]:
    """Parse and return output of the qstat command run with options
    to obtain job status.

    :param output: output of the qstat command
    :param job_id: allocation id or job step id
    :return: status
    """
    result = None
    for line in output.split("\n"):
        fields = line.split()
        if len(fields) >= 5:
            if fields[0] == job_id:
                stat = fields[4]
                result = stat
                break
    return result


def parse_qstat_jobid_json(output: str, job_id: str) -> t.Optional[str]:
    """Parse and return output of the qstat command run with JSON options
    to obtain job status.

    :param output: output of the qstat command in JSON format
    :param job_id: allocation id or job step id
    :return: status
    """
    out_json = load_and_clean_json(output)

    if "Jobs" not in out_json:
        return None
    jobs: dict[str, t.Any] = out_json["Jobs"]
    job: t.Optional[dict[str, t.Any]] = jobs.get(job_id, None)
    if job is None:
        return None
    return str(job.get("job_state", None))


def parse_qstat_nodes(output: str) -> t.List[str]:
    """Parse and return the qstat command run with
    options to obtain node list.

    This function parses and returns the nodes of
    a job in a list with the duplicates removed.

    The `output` parameter must be in JSON format.

    :param output: output of the qstat command in JSON format
    :return: compute nodes of the allocation or job
    """
    nodes: t.List[str] = []
    out_json = load_and_clean_json(output)
    if "Jobs" not in out_json:
        return nodes
    jobs = out_json["Jobs"]
    job = jobs[list(jobs.keys())[0]]
    vnodes = job["exec_vnode"]
    for vnode in vnodes.split("+"):
        vnode = vnode.strip("()")
        vnode = vnode.split(":")[0]
        nodes.append(vnode)

    return list(sorted(set(nodes)))


def parse_step_id_from_qstat(output: str, step_name: str) -> t.Optional[str]:
    """Parse and return the step id from a qstat command

    :param output: output qstat
    :param step_name: the name of the step to query
    :return: the step_id
    """
    step_id: t.Optional[str] = None
    out_json = load_and_clean_json(output)

    if "Jobs" not in out_json:
        return step_id
    jobs = out_json["Jobs"]
    for key, val in jobs.items():
        if val["Job_Name"] == step_name:
            step_id = key
            return step_id

    return step_id


def load_and_clean_json(out: str) -> t.Any:
    if len(out.strip()) == 0:
        return ""
    try:
        json_out = json.loads(out)
        return json_out
    except json.decoder.JSONDecodeError as e:
        out_arr = out.split("\n")
        new_out = ""
        for lineno, line in enumerate(out_arr):
            if lineno != e.lineno - 1:
                new_out = new_out + "\n" + line
        return load_and_clean_json(new_out)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import shutil
import subprocess
import typing as t

from ..error import LauncherError, SSUnsupportedError
from ..log import get_logger
from .base import RunSettings

logger = get_logger(__name__)


class _BaseMPISettings(RunSettings):
    """Base class for all common arguments of MPI-standard run commands"""

    def __init__(
        self,
        exe: str,
        exe_args: t.Optional[t.Union[str, t.List[str]]] = None,
        run_command: str = "mpiexec",
        run_args: t.Optional[t.Dict[str, t.Union[int, str, float, None]]] = None,
        env_vars: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        fail_if_missing_exec: bool = True,
        **kwargs: t.Any,
    ) -> None:
        """Settings to format run job with an MPI-standard binary

        Note that environment variables can be passed with a None
        value to signify that they should be exported from the current
        environment

        Any arguments passed in the ``run_args`` dict will be converted
        command line arguments and prefixed with ``--``. Values of
        None can be provided for arguments that do not have values.

        :param exe: executable
        :param exe_args: executable arguments
        :param run_args: arguments for run command
        :param env_vars: environment vars to launch job with
        :param fail_if_missing_exec: Throw an exception of the MPI command
                                     is missing. Otherwise, throw a warning
        """
        super().__init__(
            exe,
            exe_args,
            run_command=run_command,
            run_args=run_args,
            env_vars=env_vars,
            **kwargs,
        )
        self.mpmd: t.List[RunSettings] = []
        self.affinity_script: t.List[str] = []

        if not shutil.which(self._run_command):
            msg = (
                f"Cannot find {self._run_command}. Try passing the "
                "full path via run_command."
            )
            if fail_if_missing_exec:
                raise LauncherError(msg)
            logger.warning(msg)

    reserved_run_args = {"wd", "wdir"}

    def make_mpmd(self, settings: RunSettings) -> None:
        """Make a mpmd workload by combining two ``mpirun`` commands

        This connects the two settings to be executed with a single
        Model instance

        :param settings: MpirunSettings instance
        """
        if self.colocated_db_settings:
            raise SSUnsupportedError(
                "Colocated models cannot be run as a mpmd workload"
            )
        self.mpmd.append(settings)

    def set_task_map(self, task_mapping: str) -> None:
        """Set ``mpirun`` task mapping

        this sets ``--map-by <mapping>``

        For examples, see the man page for ``mpirun``

        :param task_mapping: task mapping
        """
        self.run_args["map-by"] = task_mapping

    def set_cpus_per_task(self, cpus_per_task: int) -> None:
        """Set the number of tasks for this job

        This sets ``--cpus-per-proc`` for MPI compliant implementations

        note: this option has been deprecated in openMPI 4.0+
        and will soon be replaced.

        :param cpus_per_task: number of tasks
        """
        self.run_args["cpus-per-proc"] = int(cpus_per_task)

    def set_cpu_binding_type(self, bind_type: str) -> None:
        """Specifies the cores to which MPI processes are bound

        This sets ``--bind-to`` for MPI compliant implementations

        :param bind_type: binding type
        """
        self.run_args["bind-to"] = bind_type

    def set_tasks_per_node(self, tasks_per_node: int) -> None:
        """Set the number of tasks per node

        :param tasks_per_node: number of tasks to launch per node
        """
        self.run_args["npernode"] = int(tasks_per_node)

    def set_tasks(self, tasks: int) -> None:
        """Set the number of tasks for this job

        This sets ``-n`` for MPI compliant implementations

        :param tasks: number of tasks
        """
        self.run_args["n"] = int(tasks)

    def set_hostlist(self, host_list: t.Union[str, t.List[str]]) -> None:
        """Set the hostlist for the ``mpirun`` command

        This sets ``--host``

        :param host_list: list of host names
        :raises TypeError: if not str or list of str
        """
        if isinstance(host_list, str):
            host_list = [host_list.strip()]
        if not isinstance(host_list, list):
            raise TypeError("host_list argument must be a list of strings")
        if not all(isinstance(host, str) for host in host_list):
            raise TypeError("host_list argument must be list of strings")
        self.run_args["host"] = ",".join(host_list)

    def set_hostlist_from_file(self, file_path: str) -> None:
        """Use the contents of a file to set the hostlist

        This sets ``--hostfile``

        :param file_path: Path to the hostlist file
        """
        self.run_args["hostfile"] = file_path

    def set_verbose_launch(self, verbose: bool) -> None:
        """Set the job to run in verbose mode

        This sets ``--verbose``

        :param verbose: Whether the job should be run verbosely
        """
        if verbose:
            self.run_args["verbose"] = None
        else:
            self.run_args.pop("verbose", None)

    def set_quiet_launch(self, quiet: bool) -> None:
        """Set the job to run in quiet mode

        This sets ``--quiet``

        :param quiet: Whether the job should be run quietly
        """
        if quiet:
            self.run_args["quiet"] = None
        else:
            self.run_args.pop("quiet", None)

    def set_broadcast(self, dest_path: t.Optional[str] = None) -> None:
        """Copy the specified executable(s) to remote machines

        This sets ``--preload-binary``

        :param dest_path: Destination path (Ignored)
        """
        if dest_path is not None and isinstance(dest_path, str):
            logger.warning(
                (
                    f"{type(self)} cannot set a destination path during broadcast. "
                    "Using session directory instead"
                )
            )
        self.run_args["preload-binary"] = None

    def set_walltime(self, walltime: str) -> None:
        """Set the maximum number of seconds that a job will run

        This sets ``--timeout``

        :param walltime: number like string of seconds that a job will run in secs
        """
        self.run_args["timeout"] = walltime

    def format_run_args(self) -> t.List[str]:
        """Return a list of MPI-standard formatted run arguments

        :return: list of MPI-standard arguments for these settings
        """
        # args launcher uses
        args = []
        restricted = ["wdir", "wd"]

        for opt, value in self.run_args.items():
            if opt not in restricted:
                prefix = "--"
                if not value:
                    args += [prefix + opt]
                else:
                    args += [prefix + opt, str(value)]
        return args

    def format_env_vars(self) -> t.List[str]:
        """Format the environment variables for mpirun

        :return: list of env vars
        """
        formatted = []
        env_string = "-x"

        if self.env_vars:
            for name, value in self.env_vars.items():
                if value:
                    formatted += [env_string, "=".join((name, str(value)))]
                else:
                    formatted += [env_string, name]
        return formatted


class MpirunSettings(_BaseMPISettings):
    def __init__(
        self,
        exe: str,
        exe_args: t.Optional[t.Union[str, t.List[str]]] = None,
        run_args: t.Optional[t.Dict[str, t.Union[int, str, float, None]]] = None,
        env_vars: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        **kwargs: t.Any,
    ) -> None:
        """Settings to run job with ``mpirun`` command (MPI-standard)

        Note that environment variables can be passed with a None
        value to signify that they should be exported from the current
        environment

        Any arguments passed in the ``run_args`` dict will be converted
        into ``mpirun`` arguments and prefixed with ``--``. Values of
        None can be provided for arguments that do not have values.

        :param exe: executable
        :param exe_args: executable arguments
        :param run_args: arguments for run command
        :param env_vars: environment vars to launch job with
        """
        super().__init__(exe, exe_args, "mpirun", run_args, env_vars, **kwargs)


class MpiexecSettings(_BaseMPISettings):
    def __init__(
        self,
        exe: str,
        exe_args: t.Optional[t.Union[str, t.List[str]]] = None,
        run_args: t.Optional[t.Dict[str, t.Union[int, str, float, None]]] = None,
        env_vars: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        **kwargs: t.Any,
    ) -> None:
        """Settings to run job with ``mpiexec`` command (MPI-standard)

        Note that environment variables can be passed with a None
        value to signify that they should be exported from the current
        environment

        Any arguments passed in the ``run_args`` dict will be converted
        into ``mpiexec`` arguments and prefixed with ``--``. Values of
        None can be provided for arguments that do not have values.

        :param exe: executable
        :param exe_args: executable arguments
        :param run_args: arguments for run command
        :param env_vars: environment vars to launch job with
        """
        super().__init__(exe, exe_args, "mpiexec", run_args, env_vars, **kwargs)

        completed_process = subprocess.run(
            [self._run_command, "--help"], capture_output=True, check=False
        )
        help_statement = completed_process.stdout.decode()
        if "mpiexec.slurm" in help_statement:
            raise SSUnsupportedError(
                "Slurm's wrapper for mpiexec is unsupported. Use slurmSettings instead"
            )


class OrterunSettings(_BaseMPISettings):
    def __init__(
        self,
        exe: str,
        exe_args: t.Optional[t.Union[str, t.List[str]]] = None,
        run_args: t.Optional[t.Dict[str, t.Union[int, str, float, None]]] = None,
        env_vars: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        **kwargs: t.Any,
    ) -> None:
        """Settings to run job with ``orterun`` command (MPI-standard)

        Note that environment variables can be passed with a None
        value to signify that they should be exported from the current
        environment

        Any arguments passed in the ``run_args`` dict will be converted
        into ``orterun`` arguments and prefixed with ``--``. Values of
        None can be provided for arguments that do not have values.

        :param exe: executable
        :param exe_args: executable arguments
        :param run_args: arguments for run command
        :param env_vars: environment vars to launch job with
        """
        super().__init__(exe, exe_args, "orterun", run_args, env_vars, **kwargs)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import pathlib
import typing as t
from dataclasses import dataclass, field

from ...database import Orchestrator
from ...entity import DBNode, Ensemble, EntitySequence, Model, SmartSimEntity
from ...error import SmartSimError
from ..config import CONFIG
from ..utils import helpers as _helpers
from ..utils import serialize as _serialize

_T = t.TypeVar("_T")
_U = t.TypeVar("_U")
_AtomicLaunchableT = t.TypeVar("_AtomicLaunchableT", Model, DBNode)

if t.TYPE_CHECKING:
    import os


class Manifest:
    """This class is used to keep track of all deployables generated by an
    experiment.  Different types of deployables (i.e. different
    `SmartSimEntity`-derived objects or `EntitySequence`-derived objects) can
    be accessed by using the corresponding accessor.

    Instances of ``Model``, ``Ensemble`` and ``Orchestrator``
    can all be passed as arguments
    """

    def __init__(
        self, *args: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]
    ) -> None:
        self._deployables = list(args)
        self._check_types(self._deployables)
        self._check_names(self._deployables)
        self._check_entity_lists_nonempty()

    @property
    def dbs(self) -> t.List[Orchestrator]:
        """Return a list of Orchestrator instances in Manifest

        :raises SmartSimError: if user added to databases to manifest
        :return: List of orchestrator instances
        """
        dbs = [item for item in self._deployables if isinstance(item, Orchestrator)]
        return dbs

    @property
    def models(self) -> t.List[Model]:
        """Return Model instances in Manifest

        :return: model instances
        """
        _models: t.List[Model] = [
            item for item in self._deployables if isinstance(item, Model)
        ]
        return _models

    @property
    def ensembles(self) -> t.List[Ensemble]:
        """Return Ensemble instances in Manifest

        :return: list of ensembles
        """
        return [e for e in self._deployables if isinstance(e, Ensemble)]

    @property
    def all_entity_lists(self) -> t.List[EntitySequence[SmartSimEntity]]:
        """All entity lists, including ensembles and
        exceptional ones like Orchestrator

        :return: list of entity lists
        """
        _all_entity_lists: t.List[EntitySequence[SmartSimEntity]] = list(self.ensembles)

        for db in self.dbs:
            _all_entity_lists.append(db)

        return _all_entity_lists

    @property
    def has_deployable(self) -> bool:
        """
        Return True if the manifest contains entities that
        must be physically deployed
        """
        return bool(self._deployables)

    @staticmethod
    def _check_names(deployables: t.List[t.Any]) -> None:
        used = []
        for deployable in deployables:
            name = getattr(deployable, "name", None)
            if not name:
                raise AttributeError("Entity has no name. Please set name attribute.")
            if name in used:
                raise SmartSimError("User provided two entities with the same name")
            used.append(name)

    @staticmethod
    def _check_types(deployables: t.List[t.Any]) -> None:
        for deployable in deployables:
            if not isinstance(deployable, (SmartSimEntity, EntitySequence)):
                raise TypeError(
                    f"Entity has type {type(deployable)}, not "
                    + "SmartSimEntity or EntitySequence"
                )

    def _check_entity_lists_nonempty(self) -> None:
        """Check deployables for sanity before launching"""

        for entity_list in self.all_entity_lists:
            if len(entity_list) < 1:
                raise ValueError(f"{entity_list.name} is empty. Nothing to launch.")

    def __str__(self) -> str:
        output = ""
        e_header = "=== Ensembles ===\n"
        m_header = "=== Models ===\n"
        db_header = "=== Database ===\n"
        if self.ensembles:
            output += e_header

            all_ensembles = self.ensembles
            for ensemble in all_ensembles:
                output += f"{ensemble.name}\n"
                output += f"Members: {len(ensemble)}\n"
                output += f"Batch Launch: {ensemble.batch}\n"
                if ensemble.batch:
                    output += f"{str(ensemble.batch_settings)}\n"
            output += "\n"

        if self.models:
            output += m_header
            for model in self.models:
                output += f"{model.name}\n"
                if model.batch_settings:
                    output += f"{model.batch_settings}\n"
                output += f"{model.run_settings}\n"
                if model.params:
                    output += f"Parameters: \n{_helpers.fmt_dict(model.params)}\n"
            output += "\n"

        for adb in self.dbs:
            output += db_header
            output += f"Shards: {adb.num_shards}\n"
            output += f"Port: {str(adb.ports[0])}\n"
            output += f"Network: {adb._interfaces}\n"
            output += f"Batch Launch: {adb.batch}\n"
            if adb.batch:
                output += f"{str(adb.batch_settings)}\n"

        output += "\n"
        return output

    @property
    def has_db_objects(self) -> bool:
        """Check if any entity has DBObjects to set"""
        ents: t.Iterable[t.Union[Model, Ensemble]] = itertools.chain(
            self.models,
            self.ensembles,
            (member for ens in self.ensembles for member in ens.entities),
        )
        return any(any(ent.db_models) or any(ent.db_scripts) for ent in ents)


class _LaunchedManifestMetadata(t.NamedTuple):
    run_id: str
    exp_name: str
    exp_path: str
    launcher_name: str

    @property
    def exp_telemetry_subdirectory(self) -> pathlib.Path:
        return _format_exp_telemetry_path(self.exp_path)

    @property
    def run_telemetry_subdirectory(self) -> pathlib.Path:
        return _format_run_telemetry_path(self.exp_path, self.exp_name, self.run_id)

    @property
    def manifest_file_path(self) -> pathlib.Path:
        return self.exp_telemetry_subdirectory / _serialize.MANIFEST_FILENAME


@dataclass(frozen=True)
class LaunchedManifest(t.Generic[_T]):
    """Immutable manifest mapping launched entities or collections of launched
    entities to other pieces of external data. This is commonly used to map a
    launch-able entity to its constructed ``Step`` instance without assuming
    that ``step.name == job.name`` or querying the ``JobManager`` which itself
    can be ephemeral.
    """

    metadata: _LaunchedManifestMetadata
    models: t.Tuple[t.Tuple[Model, _T], ...]
    ensembles: t.Tuple[t.Tuple[Ensemble, t.Tuple[t.Tuple[Model, _T], ...]], ...]
    databases: t.Tuple[t.Tuple[Orchestrator, t.Tuple[t.Tuple[DBNode, _T], ...]], ...]

    def map(self, func: t.Callable[[_T], _U]) -> "LaunchedManifest[_U]":
        def _map_entity_data(
            fn: t.Callable[[_T], _U],
            entity_list: t.Sequence[t.Tuple[_AtomicLaunchableT, _T]],
        ) -> t.Tuple[t.Tuple[_AtomicLaunchableT, _U], ...]:
            return tuple((entity, fn(data)) for entity, data in entity_list)

        return LaunchedManifest(
            metadata=self.metadata,
            models=_map_entity_data(func, self.models),
            ensembles=tuple(
                (ens, _map_entity_data(func, model_data))
                for ens, model_data in self.ensembles
            ),
            databases=tuple(
                (db_, _map_entity_data(func, node_data))
                for db_, node_data in self.databases
            ),
        )


@dataclass(frozen=True)
class LaunchedManifestBuilder(t.Generic[_T]):
    """A class comprised of mutable collections of SmartSim entities that is
    used to build a ``LaunchedManifest`` while going through the launching
    process.
    """

    exp_name: str
    exp_path: str
    launcher_name: str
    run_id: str = field(default_factory=_helpers.create_short_id_str)

    _models: t.List[t.Tuple[Model, _T]] = field(default_factory=list, init=False)
    _ensembles: t.List[t.Tuple[Ensemble, t.Tuple[t.Tuple[Model, _T], ...]]] = field(
        default_factory=list, init=False
    )
    _databases: t.List[t.Tuple[Orchestrator, t.Tuple[t.Tuple[DBNode, _T], ...]]] = (
        field(default_factory=list, init=False)
    )

    @property
    def exp_telemetry_subdirectory(self) -> pathlib.Path:
        return _format_exp_telemetry_path(self.exp_path)

    @property
    def run_telemetry_subdirectory(self) -> pathlib.Path:
        return _format_run_telemetry_path(self.exp_path, self.exp_name, self.run_id)

    def add_model(self, model: Model, data: _T) -> None:
        self._models.append((model, data))

    def add_ensemble(self, ens: Ensemble, data: t.Sequence[_T]) -> None:
        self._ensembles.append((ens, self._entities_to_data(ens.entities, data)))

    def add_database(self, db_: Orchestrator, data: t.Sequence[_T]) -> None:
        self._databases.append((db_, self._entities_to_data(db_.entities, data)))

    @staticmethod
    def _entities_to_data(
        entities: t.Sequence[_AtomicLaunchableT], data: t.Sequence[_T]
    ) -> t.Tuple[t.Tuple[_AtomicLaunchableT, _T], ...]:
        if not entities:
            raise ValueError("Cannot map data to an empty entity sequence")
        if len(entities) != len(data):
            raise ValueError(
                f"Cannot map data sequence of length {len(data)} to entity "
                f"sequence of length {len(entities)}"
            )
        return tuple(zip(entities, data))

    def finalize(self) -> LaunchedManifest[_T]:
        return LaunchedManifest(
            metadata=_LaunchedManifestMetadata(
                self.run_id,
                self.exp_name,
                self.exp_path,
                self.launcher_name,
            ),
            models=tuple(self._models),
            ensembles=tuple(self._ensembles),
            databases=tuple(self._databases),
        )


def _format_exp_telemetry_path(
    exp_path: t.Union[str, "os.PathLike[str]"]
) -> pathlib.Path:
    return pathlib.Path(exp_path, CONFIG.telemetry_subdir)


def _format_run_telemetry_path(
    exp_path: t.Union[str, "os.PathLike[str]"], exp_name: str, run_id: str
) -> pathlib.Path:
    return _format_exp_telemetry_path(exp_path) / f"{exp_name}/{run_id}"



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import typing as t

from smartsim._core._cli.utils import clean, get_install_path


def configure_parser(parser: argparse.ArgumentParser) -> None:
    """Builds the parser for the command"""
    parser.add_argument(
        "--clobber",
        action="store_true",
        default=False,
        help="Remove all SmartSim non-python dependencies as well",
    )


def execute(
    args: argparse.Namespace, _unparsed_args: t.Optional[t.List[str]] = None, /
) -> int:
    return clean(get_install_path() / "_core", _all=args.clobber)


def execute_all(
    args: argparse.Namespace, _unparsed_args: t.Optional[t.List[str]] = None, /
) -> int:
    args.clobber = True
    return execute(args)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import contextlib
import io
import multiprocessing as mp
import os
import os.path
import tempfile
import typing as t
from types import TracebackType

import numpy as np
from smartredis import Client

from smartsim import Experiment
from smartsim._core._cli.utils import SMART_LOGGER_FORMAT
from smartsim._core._install.builder import Device
from smartsim._core.utils.helpers import installed_redisai_backends
from smartsim._core.utils.network import find_free_port
from smartsim.log import get_logger

logger = get_logger("Smart", fmt=SMART_LOGGER_FORMAT)

# Many of the functions in this module will import optional
# ML python packages only if they are needed to test the build is working
#
# pylint: disable=import-error,import-outside-toplevel
# mypy: disable-error-code="import"


if t.TYPE_CHECKING:
    from multiprocessing.connection import Connection

    # pylint: disable-next=unsubscriptable-object
    _TemporaryDirectory = tempfile.TemporaryDirectory[str]
else:
    _TemporaryDirectory = tempfile.TemporaryDirectory


class _VerificationTempDir(_TemporaryDirectory):
    """A Temporary directory to be used as a context manager that will only
    clean itself up if no error is raised within its context
    """

    def __exit__(
        self,
        exc: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        tb: t.Optional[TracebackType],
    ) -> None:
        if not value:  # Yay, no error! Clean up as normal
            super().__exit__(exc, value, tb)
        else:  # Uh-oh! Better make sure this is not implicitly cleaned up
            self._finalizer.detach()  # type: ignore[attr-defined]


def execute(
    args: argparse.Namespace, _unparsed_args: t.Optional[t.List[str]] = None, /
) -> int:
    """Validate the SmartSim installation works as expected given a
    simple experiment
    """
    backends = installed_redisai_backends()
    temp_dir = ""
    device = Device(args.device)
    try:
        with contextlib.ExitStack() as ctx:
            temp_dir = ctx.enter_context(_VerificationTempDir(dir=os.getcwd()))
            validate_env = {
                "SR_LOG_LEVEL": os.environ.get("SR_LOG_LEVEL", "INFO"),
                "SR_LOG_FILE": os.environ.get(
                    "SR_LOG_FILE", os.path.join(temp_dir, "smartredis.log")
                ),
            }
            if device == Device.GPU:
                validate_env["CUDA_VISIBLE_DEVICES"] = "0"
            ctx.enter_context(_env_vars_set_to(validate_env))
            test_install(
                location=temp_dir,
                port=args.port,
                device=device,
                with_tf="tensorflow" in backends,
                with_pt="torch" in backends,
                with_onnx="onnxruntime" in backends,
            )
    except Exception as e:
        logger.error(
            "SmartSim failed to run a simple experiment!\n"
            f"Experiment failed due to the following exception:\n{e}",
            exc_info=True,
        )
        if temp_dir:
            logger.info(f"Output files are available at `{temp_dir}`")
        return os.EX_SOFTWARE
    return os.EX_OK


def configure_parser(parser: argparse.ArgumentParser) -> None:
    """Build the parser for the command"""
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=None,
        help=(
            "The port on which to run the orchestrator for the mini experiment. "
            "If not provided, `smart` will attempt to automatically select an "
            "open port"
        ),
    )
    parser.add_argument(
        "--device",
        type=str.lower,
        default=Device.CPU.value,
        choices=[device.value for device in Device],
        help="Device to test the ML backends against",
    )


def test_install(
    location: str,
    port: t.Optional[int],
    device: Device,
    with_tf: bool,
    with_pt: bool,
    with_onnx: bool,
) -> None:
    exp = Experiment("ValidationExperiment", exp_path=location, launcher="local")
    exp.telemetry.disable()
    port = find_free_port() if port is None else port

    with _make_managed_local_orc(exp, port) as client:
        logger.info("Verifying Tensor Transfer")
        client.put_tensor("plain-tensor", np.ones((1, 1, 3, 3)))
        client.get_tensor("plain-tensor")
        if with_pt:
            logger.info("Verifying Torch Backend")
            _test_torch_install(client, device)
        if with_onnx:
            logger.info("Verifying ONNX Backend")
            _test_onnx_install(client, device)
        if with_tf:  # Run last in case TF locks an entire GPU
            logger.info("Verifying TensorFlow Backend")
            _test_tf_install(client, location, device)
        logger.info("Success!")


@contextlib.contextmanager
def _env_vars_set_to(
    evars: t.Mapping[str, t.Optional[str]]
) -> t.Generator[None, None, None]:
    envvars = tuple((var, os.environ.pop(var, None), val) for var, val in evars.items())
    for var, _, tmpval in envvars:
        _set_or_del_env_var(var, tmpval)
    try:
        yield
    finally:
        for var, origval, _ in reversed(envvars):
            _set_or_del_env_var(var, origval)


def _set_or_del_env_var(var: str, val: t.Optional[str]) -> None:
    if val is not None:
        os.environ[var] = val
    else:
        os.environ.pop(var, None)


@contextlib.contextmanager
def _make_managed_local_orc(
    exp: Experiment, port: int
) -> t.Generator[Client, None, None]:
    """Context managed orc that will be stopped if an exception is raised"""
    orc = exp.create_database(db_nodes=1, interface="lo", port=port)
    exp.generate(orc)
    exp.start(orc)
    try:
        (client_addr,) = orc.get_address()
        yield Client(False, address=client_addr)
    finally:
        exp.stop(orc)


def _test_tf_install(client: Client, tmp_dir: str, device: Device) -> None:
    recv_conn, send_conn = mp.Pipe(duplex=False)
    # Build the model in a subproc so that keras does not hog the gpu
    proc = mp.Process(target=_build_tf_frozen_model, args=(send_conn, tmp_dir))
    proc.start()

    # do not need the sending connection in this proc anymore
    send_conn.close()

    proc.join(timeout=120)
    if proc.is_alive():
        proc.terminate()
        raise Exception("Failed to build a simple keras model within 2 minutes")
    try:
        model_path, inputs, outputs = recv_conn.recv()
    except EOFError as e:
        raise Exception(
            "Failed to receive serialized model from subprocess. "
            "Is the `tensorflow` python package installed?"
        ) from e

    client.set_model_from_file(
        "keras-fcn",
        model_path,
        "TF",
        device=device.value.upper(),
        inputs=inputs,
        outputs=outputs,
    )
    client.put_tensor("keras-input", np.random.rand(1, 28, 28).astype(np.float32))
    client.run_model("keras-fcn", inputs=["keras-input"], outputs=["keras-output"])
    client.get_tensor("keras-output")


def _build_tf_frozen_model(conn: "Connection", tmp_dir: str) -> None:
    from tensorflow import keras

    from smartsim.ml.tf import freeze_model

    fcn = keras.Sequential(
        layers=[
            keras.layers.InputLayer(input_shape=(28, 28), name="input"),
            keras.layers.Flatten(input_shape=(28, 28), name="flatten"),
            keras.layers.Dense(128, activation="relu", name="dense"),
            keras.layers.Dense(10, activation="softmax", name="output"),
        ],
        name="FullyConnectedNetwork",
    )
    fcn.compile(
        optimizer="adam", loss="sparse_categorical_crossentropy", metrics=["accuracy"]
    )
    model_path, inputs, outputs = freeze_model(fcn, tmp_dir, "keras_model.pb")
    conn.send((model_path, inputs, outputs))


def _test_torch_install(client: Client, device: Device) -> None:
    import torch
    from torch import nn

    class Net(nn.Module):
        def __init__(self) -> None:
            super().__init__()
            self.conv: t.Callable[..., torch.Tensor] = nn.Conv2d(1, 1, 3)

        def forward(self, x: torch.Tensor) -> torch.Tensor:
            return self.conv(x)

    if device == Device.GPU:
        device_ = torch.device("cuda")
    else:
        device_ = torch.device("cpu")

    net = Net()
    net.to(device_)
    net.eval()

    forward_input = torch.rand(1, 1, 3, 3).to(device_)
    traced = torch.jit.trace(net, forward_input)  # type: ignore[no-untyped-call]

    buffer = io.BytesIO()
    torch.jit.save(traced, buffer)  # type: ignore[no-untyped-call]
    model = buffer.getvalue()

    client.set_model("torch-nn", model, backend="TORCH", device=device.value.upper())
    client.put_tensor("torch-in", torch.rand(1, 1, 3, 3).numpy())
    client.run_model("torch-nn", inputs=["torch-in"], outputs=["torch-out"])
    client.get_tensor("torch-out")


def _test_onnx_install(client: Client, device: Device) -> None:
    from skl2onnx import to_onnx
    from sklearn.cluster import KMeans

    data = np.arange(20, dtype=np.float32).reshape(10, 2)
    model = KMeans(n_clusters=2, n_init=10)
    model.fit(data)

    kmeans = to_onnx(model, data, target_opset=11)
    model = kmeans.SerializeToString()
    sample = np.arange(20, dtype=np.float32).reshape(10, 2)

    client.put_tensor("onnx-input", sample)
    client.set_model("onnx-kmeans", model, "ONNX", device=device.value.upper())
    client.run_model(
        "onnx-kmeans", inputs=["onnx-input"], outputs=["onnx-labels", "onnx-transform"]
    )
    client.get_tensor("onnx-labels")



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pathlib
import shutil
import typing as t
from datetime import datetime
from distutils import dir_util  # pylint: disable=deprecated-module
from logging import DEBUG, INFO
from os import mkdir, path, symlink
from os.path import join, relpath

from tabulate import tabulate

from ...database import Orchestrator
from ...entity import Ensemble, Model, TaggedFilesHierarchy
from ...log import get_logger
from ..control import Manifest
from .modelwriter import ModelWriter

logger = get_logger(__name__)
logger.propagate = False


class Generator:
    """The primary job of the generator is to create the file structure
    for a SmartSim experiment. The Generator is responsible for reading
    and writing into configuration files as well.
    """

    def __init__(
        self, gen_path: str, overwrite: bool = False, verbose: bool = True
    ) -> None:
        """Initialize a generator object

        if overwrite is true, replace any existing
        configured models within an ensemble if there
        is a name collision. Also replace any and all directories
        for the experiment with fresh copies. Otherwise, if overwrite
        is false, raises EntityExistsError when there is a name
        collision between entities.

        :param gen_path: Path in which files need to be generated
        :param overwrite: toggle entity replacement
        :param verbose: Whether generation information should be logged to std out
        """
        self._writer = ModelWriter()
        self.gen_path = gen_path
        self.overwrite = overwrite
        self.log_level = DEBUG if not verbose else INFO

    @property
    def log_file(self) -> str:
        """Returns the location of the file
        summarizing the parameters used for the last generation
        of all generated entities.

        :returns: path to file with parameter settings
        """
        return join(self.gen_path, "smartsim_params.txt")

    def generate_experiment(self, *args: t.Any) -> None:
        """Run ensemble and experiment file structure generation

        Generate the file structure for a SmartSim experiment. This
        includes the writing and configuring of input files for a
        model.

        To have files or directories present in the created entity
        directories, such as datasets or input files, call
        ``entity.attach_generator_files`` prior to generation. See
        ``entity.attach_generator_files`` for more information on
        what types of files can be included.

        Tagged model files are read, checked for input variables to
        configure, and written. Input variables to configure are
        specified with a tag within the input file itself.
        The default tag is surronding an input value with semicolons.
        e.g. ``THERMO=;90;``

        """
        generator_manifest = Manifest(*args)

        self._gen_exp_dir()
        self._gen_orc_dir(generator_manifest.dbs)
        self._gen_entity_list_dir(generator_manifest.ensembles)
        self._gen_entity_dirs(generator_manifest.models)

    def set_tag(self, tag: str, regex: t.Optional[str] = None) -> None:
        """Set the tag used for tagging input files

        Set a tag or a regular expression for the
        generator to look for when configuring new models.

        For example, a tag might be ``;`` where the
        expression being replaced in the model configuration
        file would look like ``;expression;``

        A full regular expression might tag specific
        model configurations such that the configuration
        files don't need to be tagged manually.

        :param tag: A string of characters that signify
                    the string to be changed. Defaults to ``;``
        :param regex: full regex for the modelwriter to search for
        """
        self._writer.set_tag(tag, regex)

    def _gen_exp_dir(self) -> None:
        """Create the directory for an experiment if it does not
        already exist.
        """

        if path.isfile(self.gen_path):
            raise FileExistsError(
                f"Experiment directory could not be created. {self.gen_path} exists"
            )
        if not path.isdir(self.gen_path):
            # keep exists ok for race conditions on NFS
            pathlib.Path(self.gen_path).mkdir(exist_ok=True, parents=True)
        else:
            logger.log(
                level=self.log_level, msg="Working in previously created experiment"
            )

        # The log_file only keeps track of the last generation
        # this is to avoid gigantic files in case the user repeats
        # generation several times. The information is anyhow
        # redundant, as it is also written in each entity's dir
        with open(self.log_file, mode="w", encoding="utf-8") as log_file:
            dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
            log_file.write(f"Generation start date and time: {dt_string}\n")

    def _gen_orc_dir(self, orchestrator_list: t.List[Orchestrator]) -> None:
        """Create the directory that will hold the error, output and
           configuration files for the orchestrator.

        :param orchestrator: Orchestrator instance
        """
        # Loop through orchestrators
        for orchestrator in orchestrator_list:
            orc_path = path.join(self.gen_path, orchestrator.name)

            orchestrator.set_path(orc_path)
            # Always remove orchestrator files if present.
            if path.isdir(orc_path):
                shutil.rmtree(orc_path, ignore_errors=True)
            pathlib.Path(orc_path).mkdir(exist_ok=self.overwrite, parents=True)

    def _gen_entity_list_dir(self, entity_lists: t.List[Ensemble]) -> None:
        """Generate directories for Ensemble instances

        :param entity_lists: list of Ensemble instances
        """

        if not entity_lists:
            return

        for elist in entity_lists:
            elist_dir = path.join(self.gen_path, elist.name)
            if path.isdir(elist_dir):
                if self.overwrite:
                    shutil.rmtree(elist_dir)
                    mkdir(elist_dir)
            else:
                mkdir(elist_dir)
            elist.path = elist_dir

            self._gen_entity_dirs(list(elist.models), entity_list=elist)

    def _gen_entity_dirs(
        self,
        entities: t.List[Model],
        entity_list: t.Optional[Ensemble] = None,
    ) -> None:
        """Generate directories for Entity instances

        :param entities: list of Model instances
        :param entity_list: Ensemble instance
        :raises EntityExistsError: if a directory already exists for an
                                   entity by that name
        """
        if not entities:
            return

        for entity in entities:
            if entity_list:
                dst = path.join(self.gen_path, entity_list.name, entity.name)
            else:
                dst = path.join(self.gen_path, entity.name)

            if path.isdir(dst):
                if self.overwrite:
                    shutil.rmtree(dst)
                else:
                    error = (
                        f"Directory for entity {entity.name} "
                        f"already exists in path {dst}"
                    )
                    raise FileExistsError(error)
            pathlib.Path(dst).mkdir(exist_ok=True)
            entity.path = dst

            self._copy_entity_files(entity)
            self._link_entity_files(entity)
            self._write_tagged_entity_files(entity)

    def _write_tagged_entity_files(self, entity: Model) -> None:
        """Read, configure and write the tagged input files for
           a Model instance within an ensemble. This function
           specifically deals with the tagged files attached to
           an Ensemble.

        :param entity: a Model instance
        """
        if entity.files:
            to_write = []

            def _build_tagged_files(tagged: TaggedFilesHierarchy) -> None:
                """Using a TaggedFileHierarchy, reproduce the tagged file
                directory structure

                :param tagged: a TaggedFileHierarchy to be built as a
                               directory structure
                """
                for file in tagged.files:
                    dst_path = path.join(entity.path, tagged.base, path.basename(file))
                    shutil.copyfile(file, dst_path)
                    to_write.append(dst_path)

                for tagged_dir in tagged.dirs:
                    mkdir(
                        path.join(
                            entity.path, tagged.base, path.basename(tagged_dir.base)
                        )
                    )
                    _build_tagged_files(tagged_dir)

            if entity.files.tagged_hierarchy:
                _build_tagged_files(entity.files.tagged_hierarchy)

            # write in changes to configurations
            if isinstance(entity, Model):
                files_to_params = self._writer.configure_tagged_model_files(
                    to_write, entity.params
                )
                self._log_params(entity, files_to_params)

    def _log_params(
        self, entity: Model, files_to_params: t.Dict[str, t.Dict[str, str]]
    ) -> None:
        """Log which files were modified during generation

        and what values were set to the parameters

        :param entity: the model being generated
        :param files_to_params: a dict connecting each file to its parameter settings
        """
        used_params: t.Dict[str, str] = {}
        file_to_tables: t.Dict[str, str] = {}
        for file, params in files_to_params.items():
            used_params.update(params)
            table = tabulate(params.items(), headers=["Name", "Value"])
            file_to_tables[relpath(file, self.gen_path)] = table

        if used_params:
            used_params_str = ", ".join(
                [f"{name}={value}" for name, value in used_params.items()]
            )
            logger.log(
                level=self.log_level,
                msg=f"Configured model {entity.name} with params {used_params_str}",
            )
            file_table = tabulate(
                file_to_tables.items(),
                headers=["File name", "Parameters"],
            )
            log_entry = f"Model name: {entity.name}\n{file_table}\n\n"
            with open(self.log_file, mode="a", encoding="utf-8") as logfile:
                logfile.write(log_entry)
            with open(
                join(entity.path, "smartsim_params.txt"), mode="w", encoding="utf-8"
            ) as local_logfile:
                local_logfile.write(log_entry)

        else:
            logger.log(
                level=self.log_level,
                msg=f"Configured model {entity.name} with no parameters",
            )

    @staticmethod
    def _copy_entity_files(entity: Model) -> None:
        """Copy the entity files and directories attached to this entity.

        :param entity: Model
        """
        if entity.files:
            for to_copy in entity.files.copy:
                dst_path = path.join(entity.path, path.basename(to_copy))
                if path.isdir(to_copy):
                    dir_util.copy_tree(to_copy, entity.path)
                else:
                    shutil.copyfile(to_copy, dst_path)

    @staticmethod
    def _link_entity_files(entity: Model) -> None:
        """Symlink the entity files attached to this entity.

        :param entity: Model
        """
        if entity.files:
            for to_link in entity.files.link:
                dst_path = path.join(entity.path, path.basename(to_link))
                symlink(to_link, dst_path)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

from ..error import SSConfigError
from ..log import get_logger
from .base import BatchSettings

logger = get_logger(__name__)


class QsubBatchSettings(BatchSettings):
    def __init__(
        self,
        nodes: t.Optional[int] = None,
        ncpus: t.Optional[int] = None,
        time: t.Optional[str] = None,
        queue: t.Optional[str] = None,
        account: t.Optional[str] = None,
        resources: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
        batch_args: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        **kwargs: t.Any,
    ):
        """Specify ``qsub`` batch parameters for a job

        ``nodes``, and ``ncpus`` are used to create the
        select statement for PBS if a select statement is not
        included in the ``resources``. If both are supplied
        the value for select statement supplied in ``resources``
        will override.

        :param nodes: number of nodes for batch
        :param ncpus: number of cpus per node
        :param time: walltime for batch job
        :param queue: queue to run batch in
        :param account: account for batch launch
        :param resources: overrides for resource arguments
        :param batch_args: overrides for PBS batch arguments
        """

        self._ncpus = ncpus

        self.resources = resources or {}
        resource_nodes = self.resources.get("nodes", None)

        if nodes and resource_nodes:
            raise ValueError(
                "nodes was incorrectly specified as a constructor parameter and also "
                "as a key in the resource mapping"
            )

        # time, queue, nodes, and account set in parent class init
        super().__init__(
            "qsub",
            batch_args=batch_args,
            nodes=nodes,
            account=account,
            queue=queue,
            time=time,
            **kwargs,
        )

        self._hosts: t.List[str] = []

    @property
    def resources(self) -> t.Dict[str, t.Union[str, int]]:
        return self._resources.copy()

    @resources.setter
    def resources(self, resources: t.Dict[str, t.Union[str, int]]) -> None:
        self._sanity_check_resources(resources)
        self._resources = resources.copy()

    def set_nodes(self, num_nodes: int) -> None:
        """Set the number of nodes for this batch job

        In PBS, 'select' is the more primitive way of describing how
        many nodes to allocate for the job. 'nodes' is equivalent to
        'select' with a 'place' statement. Assuming that only advanced
        users would use 'set_resource' instead, defining the number of
        nodes here is sets the 'nodes' resource.

        :param num_nodes: number of nodes
        """

        if num_nodes:
            self.set_resource("nodes", num_nodes)

    def set_hostlist(self, host_list: t.Union[str, t.List[str]]) -> None:
        """Specify the hostlist for this job

        :param host_list: hosts to launch on
        :raises TypeError: if not str or list of str
        """
        if isinstance(host_list, str):
            host_list = [host_list.strip()]
        if not isinstance(host_list, list):
            raise TypeError("host_list argument must be a list of strings")
        if not all(isinstance(host, str) for host in host_list):
            raise TypeError("host_list argument must be a list of strings")
        self._hosts = host_list

    def set_walltime(self, walltime: str) -> None:
        """Set the walltime of the job

        format = "HH:MM:SS"

        If a walltime argument is provided in
        ``QsubBatchSettings.resources``, then
        this value will be overridden

        :param walltime: wall time
        """
        if walltime:
            self.set_resource("walltime", walltime)

    def set_queue(self, queue: str) -> None:
        """Set the queue for the batch job

        :param queue: queue name
        """
        if queue:
            self.batch_args["q"] = str(queue)

    def set_ncpus(self, num_cpus: t.Union[int, str]) -> None:
        """Set the number of cpus obtained in each node.

        If a select argument is provided in
        ``QsubBatchSettings.resources``, then
        this value will be overridden

        :param num_cpus: number of cpus per node in select
        """
        self._ncpus = int(num_cpus)

    def set_account(self, account: str) -> None:
        """Set the account for this batch job

        :param acct: account id
        """
        if account:
            self.batch_args["A"] = str(account)

    def set_resource(self, resource_name: str, value: t.Union[str, int]) -> None:
        """Set a resource value for the Qsub batch

        If a select statement is provided, the nodes and ncpus
        arguments will be overridden. Likewise for Walltime

        :param resource_name: name of resource, e.g. walltime
        :param value: value
        """
        # TODO add error checking here
        # TODO include option to overwrite place (warning for orchestrator?)
        updated_dict = self.resources
        updated_dict.update({resource_name: value})
        self._sanity_check_resources(updated_dict)
        self.resources = updated_dict

    def format_batch_args(self) -> t.List[str]:
        """Get the formatted batch arguments for a preview

        :return: batch arguments for Qsub
        :raises ValueError: if options are supplied without values
        """
        opts = self._create_resource_list()
        for opt, value in self.batch_args.items():
            prefix = "-"
            if not value:
                raise ValueError("PBS options without values are not allowed")
            opts += [" ".join((prefix + opt, str(value)))]
        return opts

    def _sanity_check_resources(
        self, resources: t.Optional[t.Dict[str, t.Union[str, int]]] = None
    ) -> None:
        """Check that only select or nodes was specified in resources

        Note: For PBS Pro, nodes is equivalent to 'select' and 'place' so
        they are not quite synonyms. Here we assume that
        """
        # Note: isinstance check here to avoid collision with default
        checked_resources = resources if isinstance(resources, dict) else self.resources

        has_select = checked_resources.get("select", None)
        has_nodes = checked_resources.get("nodes", None)

        if has_select and has_nodes:
            raise SSConfigError(
                "'select' and 'nodes' cannot both be specified. This can happen "
                "if nodes were specified using the 'set_nodes' method and "
                "'select' was set using 'set_resource'. Please only specify one."
            )

        if has_select and not isinstance(has_select, int):
            raise TypeError("The value for 'select' must be an integer")
        if has_nodes and not isinstance(has_nodes, int):
            raise TypeError("The value for 'nodes' must be an integer")

        for key, value in checked_resources.items():
            if not isinstance(key, str):
                raise TypeError(
                    f"The type of {key=} is {type(key)}. Only int and str "
                    "are allowed."
                )
            if not isinstance(value, (str, int)):
                raise TypeError(
                    f"The value associated with {key=} is {type(value)}. Only int "
                    "and str are allowed."
                )

    def _create_resource_list(self) -> t.List[str]:
        self._sanity_check_resources()
        res = []

        # Pop off some specific keywords that need to be treated separately
        resources = self.resources  # Note this is a copy so not modifying original

        # Construct the basic select/nodes statement
        if select := resources.pop("select", None):
            select_command = f"-l select={select}"
        elif nodes := resources.pop("nodes", None):
            select_command = f"-l nodes={nodes}"
        else:
            raise SSConfigError(
                "Insufficient resource specification: no nodes or select statement"
            )
        if self._ncpus:
            select_command += f":ncpus={self._ncpus}"
        if self._hosts:
            hosts = ["=".join(("host", str(host))) for host in self._hosts]
            select_command += f":{'+'.join(hosts)}"
        res += [select_command]

        # All other "standard" resource specs
        for resource, value in resources.items():
            res += [f"-l {resource}={value}"]

        return res



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import os
import typing as t

from ...._core.launcher.stepMapping import StepMap
from ....error import LauncherError, SmartSimError
from ....log import get_logger
from ....settings import (
    DragonRunSettings,
    QsubBatchSettings,
    RunSettings,
    SbatchSettings,
    SettingsBase,
)
from ....status import SmartSimStatus
from ...schemas import (
    DragonRunRequest,
    DragonRunResponse,
    DragonStopRequest,
    DragonStopResponse,
    DragonUpdateStatusRequest,
    DragonUpdateStatusResponse,
)
from ..launcher import WLMLauncher
from ..pbs.pbsLauncher import PBSLauncher
from ..slurm.slurmLauncher import SlurmLauncher
from ..step import DragonBatchStep, DragonStep, LocalStep, Step
from ..stepInfo import StepInfo
from .dragonConnector import DragonConnector, _SchemaT

logger = get_logger(__name__)


class DragonLauncher(WLMLauncher):
    """This class encapsulates the functionality needed
    to launch jobs on systems that use Dragon on top of a workload manager.

    All WLM launchers are capable of launching managed and unmanaged
    jobs. Managed jobs are queried through interaction with with WLM,
    in this case the Dragon server. Unmanaged jobs are held in the TaskManager
    and are managed through references to their launching process ID
    i.e. a psutil.Popen object.
    Batch Jobs are routed to either Slurm or PBS and their step ID
    is stored, prefixed with the name of the scheduler, to allow
    the Job Manager to interact with it.
    """

    def __init__(self) -> None:
        super().__init__()
        self._connector = DragonConnector()
        """Connector used to start and interact with the Dragon server"""
        self._slurm_launcher = SlurmLauncher()
        """Slurm sub-launcher, used only for batch jobs"""
        self._pbs_launcher = PBSLauncher()
        """PBS sub-launcher, used only for batch jobs"""

    @property
    def is_connected(self) -> bool:
        return self._connector.is_connected

    def cleanup(self) -> None:
        self._connector.cleanup()

    # RunSettings types supported by this launcher
    @property
    def supported_rs(self) -> t.Dict[t.Type[SettingsBase], t.Type[Step]]:
        # RunSettings types supported by this launcher
        return {
            DragonRunSettings: DragonStep,
            SbatchSettings: DragonBatchStep,
            QsubBatchSettings: DragonBatchStep,
            RunSettings: LocalStep,
        }

    def add_step_to_mapping_table(self, name: str, step_map: StepMap) -> None:
        super().add_step_to_mapping_table(name, step_map)

        if step_map.step_id is None:
            return
        sublauncher: t.Optional[t.Union[SlurmLauncher, PBSLauncher]] = None
        if step_map.step_id.startswith("SLURM-"):
            sublauncher = self._slurm_launcher
        elif step_map.step_id.startswith("PBS-"):
            sublauncher = self._pbs_launcher
        else:
            return

        sublauncher_step_map = StepMap(
            step_id=DragonLauncher._unprefix_step_id(step_map.step_id),
            task_id=step_map.task_id,
            managed=step_map.managed,
        )
        sublauncher.add_step_to_mapping_table(name, sublauncher_step_map)

    def run(self, step: Step) -> t.Optional[str]:
        """Run a job step through Slurm

        :param step: a job step instance
        :raises LauncherError: if launch fails
        :return: job step id if job is managed
        """

        if not self.task_manager.actively_monitoring:
            self.task_manager.start()

        step_id = None
        task_id = None

        cmd = step.get_launch_cmd()
        out, err = step.get_output_files()

        if isinstance(step, DragonBatchStep):
            # wait for batch step to submit successfully
            sublauncher_step_id: t.Optional[str] = None
            return_code, out, err = self.task_manager.start_and_wait(cmd, step.cwd)
            if return_code != 0:
                raise LauncherError(f"Sbatch submission failed\n {out}\n {err}")
            if out:
                sublauncher_step_id = out.strip()
                logger.debug(
                    f"Gleaned batch job id: {sublauncher_step_id} for {step.name}"
                )

            if sublauncher_step_id is None:
                raise SmartSimError("Could not get step id for batch step")

            if isinstance(step.batch_settings, SbatchSettings):
                self._slurm_launcher.step_mapping.add(
                    step.name, sublauncher_step_id, task_id, step.managed
                )
                step_id = "SLURM-" + sublauncher_step_id
            elif isinstance(step.batch_settings, QsubBatchSettings):
                self._pbs_launcher.step_mapping.add(
                    step.name, sublauncher_step_id, task_id, step.managed
                )
                step_id = "PBS-" + sublauncher_step_id
        elif isinstance(step, DragonStep):
            run_args = step.run_settings.run_args
            req_env = step.run_settings.env_vars
            self._connector.load_persisted_env()
            merged_env = self._connector.merge_persisted_env(os.environ.copy())
            nodes = int(run_args.get("nodes", None) or 1)
            tasks_per_node = int(run_args.get("tasks-per-node", None) or 1)
            response = _assert_schema_type(
                self._connector.send_request(
                    DragonRunRequest(
                        exe=cmd[0],
                        exe_args=cmd[1:],
                        path=step.cwd,
                        name=step.name,
                        nodes=nodes,
                        tasks_per_node=tasks_per_node,
                        env=req_env,
                        current_env=merged_env,
                        output_file=out,
                        error_file=err,
                    )
                ),
                DragonRunResponse,
            )
            step_id = str(response.step_id)
        else:
            # pylint: disable-next=consider-using-with
            out_strm = open(out, "w+", encoding="utf-8")
            # pylint: disable-next=consider-using-with
            err_strm = open(err, "w+", encoding="utf-8")
            task_id = self.task_manager.start_task(
                cmd, step.cwd, step.env, out=out_strm.fileno(), err=err_strm.fileno()
            )
            step.managed = False

        self.step_mapping.add(step.name, step_id, task_id, step.managed)

        return step_id

    def stop(self, step_name: str) -> StepInfo:
        """Step a job step

        :param step_name: name of the job to stop
        :return: update for job due to cancel
        """

        stepmap = self.step_mapping[step_name]
        step_id = str(stepmap.step_id)

        if step_id.startswith("SLURM-"):
            return self._slurm_launcher.stop(step_name)

        if step_id.startswith("PBS-"):
            return self._pbs_launcher.stop(step_name)

        _assert_schema_type(
            self._connector.send_request(DragonStopRequest(step_id=step_id)),
            DragonStopResponse,
        )

        _, step_info = self.get_step_update([step_name])[0]
        if not step_info:
            raise LauncherError(f"Could not get step_info for job step {step_name}")

        step_info.status = (
            SmartSimStatus.STATUS_CANCELLED  # set status to cancelled instead of failed
        )
        step_info.launcher_status = str(SmartSimStatus.STATUS_CANCELLED)
        return step_info

    @staticmethod
    def _unprefix_step_id(step_id: str) -> str:
        return step_id.split("-", maxsplit=1)[1]

    def _get_managed_step_update(self, step_ids: t.List[str]) -> t.List[StepInfo]:
        """Get step updates for Dragon-managed jobs

        :param step_ids: list of job step ids
        :return: list of updates for managed jobs
        """

        step_id_updates: dict[str, StepInfo] = {}

        dragon_step_ids: t.List[str] = []
        slurm_step_ids: t.List[str] = []
        pbs_step_ids: t.List[str] = []
        for step_id in step_ids:
            if step_id.startswith("SLURM-"):
                slurm_step_ids.append(step_id)
            elif step_id.startswith("PBS-"):
                pbs_step_ids.append(step_id)
            else:
                dragon_step_ids.append(step_id)

        if slurm_step_ids:
            # pylint: disable-next=protected-access
            slurm_updates = self._slurm_launcher._get_managed_step_update(
                [
                    DragonLauncher._unprefix_step_id(step_id)
                    for step_id in slurm_step_ids
                ]
            )
            step_id_updates.update(dict(zip(slurm_step_ids, slurm_updates)))

        if pbs_step_ids:
            # pylint: disable-next=protected-access
            pbs_updates = self._pbs_launcher._get_managed_step_update(
                [DragonLauncher._unprefix_step_id(step_id) for step_id in pbs_step_ids]
            )
            step_id_updates.update(dict(zip(pbs_step_ids, pbs_updates)))

        if dragon_step_ids:
            response = _assert_schema_type(
                self._connector.send_request(
                    DragonUpdateStatusRequest(step_ids=dragon_step_ids)
                ),
                DragonUpdateStatusResponse,
            )

            for step_id in step_ids:
                if step_id not in response.statuses:
                    msg = "Missing step id update from Dragon launcher."
                    if response.error_message is not None:
                        msg += "\nDragon backend reported following error: "
                        msg += response.error_message
                    logger.error(msg)
                    info = StepInfo(
                        SmartSimStatus.STATUS_FAILED,
                        SmartSimStatus.STATUS_FAILED.value,
                        -1,
                    )
                else:
                    status, ret_codes = response.statuses[step_id]
                    if ret_codes:
                        grp_ret_code = min(ret_codes)
                        if any(ret_codes):
                            _err_msg = (
                                f"One or more processes failed for job {step_id} "
                                f"Return codes were: {ret_codes}"
                            )
                            logger.error(_err_msg)
                    else:
                        grp_ret_code = None
                    info = StepInfo(status, status.value, grp_ret_code)

                step_id_updates[step_id] = info

        # Order matters as we return an ordered list of StepInfo objects
        return [step_id_updates[step_id] for step_id in step_ids]

    def __str__(self) -> str:
        return "Dragon"


def _assert_schema_type(obj: object, typ: t.Type[_SchemaT], /) -> _SchemaT:
    if not isinstance(obj, typ):
        raise TypeError(f"Expected schema of type `{typ}`, but got {type(obj)}")
    return obj
